from pywapor.collect.Globcover.DataAccess import DownloadData
import sys
from pywapor.general.logger import log

def main(Dir, latlim, lonlim, Waitbar = 1, **kwargs):
    """
    Downloads Globcover data from http://due.esrin.esa.int/page_globcover.php

    The following keyword arguments are needed:
    Dir -- 'C:/file/to/path/'
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]

    """
    log.info("--> Downloading GLOBCOVER.")

    # Download and process the data
    outpath = DownloadData(Dir, latlim, lonlim)

    return outpath

if __name__ == '__main__':
    main(sys.argv)



# -*- coding: utf-8 -*-
import sys
from pywapor.collect.MERRA2.DataAccess import DownloadData
import pywapor
from pywapor.general.logger import log
import numpy as np

def main(Dir, latlim, lonlim, Startdate, Enddate, Vars, Periods = list(range(1, 25)), Waitbar = False, verbose = True):
    """
    This function downloads MERRA inst data for a given variable, time
    interval, and spatial extent.

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Vars -- ['t2m', 'v2m'] 
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
	Periods -- [1,2,3,4,5,6,7,8,23,24] Period that needs to be downloaded. 1 period is 1 hour starting from noon
    Waitbar -- 1 (Default) Will print a waitbar
    """

    username, password = pywapor.collect.accounts.get("NASA")

    all_files = list()

    if not isinstance(Startdate, list):
        Startdate = [Startdate]
    if not isinstance(Enddate, list):
        Startdate = [Enddate]

    if not isinstance(Periods[0], list):
        Periods = [Periods for _ in Startdate]
    
    if not verbose:
        log.info("--> Downloading MERRA2 (hourly).")
    
    for Var in Vars:

        for sd, ed, periods in zip(Startdate, Enddate, Periods):

            for period in periods:

                # Download data
                files = DownloadData(Dir, Var, sd, ed, latlim, lonlim, "hourly_MERRA2", period, username, password, Waitbar, data_type = ["mean"])
                all_files = all_files + files

    all_files = np.unique(all_files).tolist()

    return all_files

if __name__ == '__main__':
    main(sys.argv)





import os
from pywapor.collect.SRTM.DataAccess import DownloadData
import sys
from pywapor.general.logger import log

def main(Dir, latlim, lonlim, **kwargs):
    """
    Downloads HydroSHED data from http://srtm.csi.cgiar.org/download

    this data includes a Digital Elevation Model (DEM)
    The spatial resolution is 90m (3s)

    The following keyword arguments are needed:
    Dir -- 'C:/file/to/path/'
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
    """

    log.info("--> Downloading SRTM.")

    # Create directory if not exists for the output
    output_folder = os.path.join(Dir, 'SRTM', 'DEM')
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

    # Define the output map and create this if not exists
    nameEnd = os.path.join(output_folder, 'DEM_SRTM_m_-_2000.02.01.00.00.tif')

    if not os.path.exists(nameEnd):

        # Download and process the data
        DownloadData(output_folder, latlim, lonlim)

    return nameEnd

if __name__ == '__main__':
    main(sys.argv)



# -*- coding: utf-8 -*-
"""

This module contains scripts used to download Level 1 data (data directly from web).

Products                      Dates                             Password
CHIRPS (daily)                1981/01/01-now                    -
SRTM                           -                                 -
GLDAS                         2000/01/01-now                    NASA
MCD43 (daily)                 2000/02/24-now                    NASA
MOD11 (daily)                 2000/02/24-now                    NASA
MYD11 (daily)                 2000/02/24-now                    NASA
MYD13 (16-daily)              2000/02/18-now                    NASA
MOD13 (16-daily)              2000/02/18-now                    NASA
MERRA
MSGCPP
GEOS

Examples:
from pyWAPOR import Collect
help(Collect)
dir(Collect)
"""

from pywapor.collect import CHIRPS, MOD11, MYD11, MCD43, MOD13, MYD13, MERRA2, Globcover, GEOS5, SRTM, WAPOR, PROBAV, STATICS, accounts, Landsat, sideloader, downloader

__all__ = ['CHIRPS', 'MOD11', 'MYD11', 'MCD43', 'MOD13', 'MYD13', 'MERRA2', 'SRTM', 'Globcover', 'GEOS5', 'WAPOR' , 'PROBAV', 'STATICS', 'accounts', 'Landsat', 'sideloader', 'downloader']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""Module with a range of general purpose GIS tools, mainly (but not only) consisting of
GDAL routines. 

Some functions in this module will be deprecated in future versions.
"""
import os
from osgeo import osr
from osgeo import gdal
import gzip
import zipfile
import numpy as np
from scipy.interpolate import NearestNDInterpolator, LinearNDInterpolator
from pywapor.general.logger import log
import xarray as xr
import pandas as pd

def ds_remove_except(ds, keep_vars):
    """Remove all variables from a dataset except the variables specified with
    `keep_vars`. Variables that are coordinates are never removed.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset from which to remove variables.
    keep_vars : list
        List of variables to keep.

    Returns
    -------
    xr.Dataset
        Dataset from which variables have been removed.
    """
    keep_vars_coords = list()
    for var in keep_vars:
        if var in list(ds.variables):
            keep_vars_coords += [k for k, v in ds[var].coords.items() if v.size > 0]
    keep_vars_coords = keep_vars + np.unique(keep_vars_coords).tolist()
    drop_vars = [x for x in list(ds.variables) if x not in keep_vars_coords]
    ds = ds.drop_vars(drop_vars)
    return ds

def export_ds_to_tif(ds, vars, base_folder):
    """Export selected `vars` from a xr.Dataset (`ds`) into `base_folder` as geotiffs.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to export.
    vars : list
        Variables in `ds` to export.
    base_folder : str
        Path to folder in which to store geotiffs.

    Returns
    -------
    dict
        Keys are variable names, values are lists with paths to the generated
        geotiffs.
    """

    all_files = dict()

    if isinstance(base_folder, type(None)):
        base_folder = os.path.split(ds.encoding["source"])[0]
    
    for var in vars:

        if var not in list(ds.variables):
            continue

        if "lat" not in list(ds[var].coords) or "lon" not in list(ds[var].coords):
            continue

        all_files[var] = list()

        folder = os.path.join(base_folder, f"{var}_output")
        
        if not os.path.exists(folder):
            os.makedirs(folder)

        if "source" in ds[var].attrs.keys():
            source = ds[var].source
        else:
            source = "-"

        if "unit" in ds[var].attrs.keys():
            unit = ds[var].unit
        else:
            unit = "-"

        if "time" in list(ds[var].coords):
            iterator = ds.time
            instantaneous = True
        elif "epoch" in list(ds[var].coords):
            iterator = ds.epoch
            instantaneous = False
        else:
            log.warning(f"! --> not exporting {var}")

        for t in iterator:

            if instantaneous:
                tlength = "inst"
                date_str = pd.Timestamp(t.values).strftime("%Y.%m.%d.%H.%M")
                array = ds[var].sel(time = t).values
            else:
                tdelta = ds.epoch_ends.sel(epoch = t) - ds.epoch_starts.sel(epoch = t)
                tlength = tdelta.values.astype('timedelta64[D]').astype(int)
                date_str = pd.Timestamp(ds.epoch_starts.sel(epoch = t).values).strftime("%Y.%m.%d")
                array = ds[var].sel(epoch = t).values

            fn = f"{var.replace('_','-')}_{source}_{unit}_{tlength}_{date_str}.tif"
            fh = os.path.join(folder, fn)

            Save_as_tiff(fh, array, ds.geotransform, ds.projection)

            all_files[var].append(fh)

    return all_files

def select_template(fhs):
    """Given a list of lists with paths to geotiffs, determines which one of 
    all the geotiffs has the highest resolution. Note that this is not based on 
    pixel size, so passing geotiffs with very different domains can give unexpected
    results.

    Parameters
    ----------
    fhs : list
        List of lists with paths to geotiffs.

    Returns
    -------
    str
        First file with the highest resolution.
    xr.Dataset
        Dataset with the required `lat` and `lon` values.
    tuple
        The geotransform of the selected geotiff.
    resolution
        Pixel size in units of the geotiffs projection.
    """
    # Flatten a list-of-lists into a  single list.
    fhs = [val for sublist in fhs for val in sublist]

    # Determine amount of pixels in each file.
    sizes = [gdal.Open(fh).RasterXSize * gdal.Open(fh).RasterYSize for fh in fhs]

    # Find index of file with most pixels.
    idx = np.argmax(sizes)

    # Create resample info.
    example_fh = fhs[idx]
    example_ds = xr.open_dataset(example_fh).isel(band = 0).drop_vars(["band", "spatial_ref"]).rename({"x": "lon", "y": "lat"})
    example_geoinfo = get_geoinfo(example_fh)

    # Calculate pixel size in meters.
    resolution = get_resolution(example_fh)
    log.info(f"--> Resampling resolution is ~{resolution:.0f} meter.")

    return example_fh, example_ds, example_geoinfo, resolution

def get_resolution(example_filepath):
    """Get the average resolution of the given geotiff.

    Parameters
    ----------
    example_filepath : str
        Path to geotiff.

    Returns
    -------
    float
        Pixel size in units of the geotiffs projection.
    """
    ds = gdal.Open(example_filepath)
    geo_ex = ds.GetGeoTransform()
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    dlat, dlon = calc_dlat_dlon(geo_ex, xsize, ysize)
    dem_resolution = (np.nanmean(dlon) + np.nanmean(dlat))/2
    return dem_resolution

def reproject_clip(source_fp, dest_fp = None, bb = None,
                    compress = True, dstSRS = "epsg:4326"):
    """Function that calls gdal.Warp(), can be used to clip and reproject 
    a geotiff in one go.

    Parameters
    ----------
    source_fp : str
        Path to file to be warped.
    dest_fp : [type], optional
        Warped output file, overwrites input when `None` given, by default None.
    bb : tuple, optional
        Clip the geotiff the a bounding-box.
        First item is latlim, second item is lonlim, by default None.
    compress : bool, optional
        Compress the data using DEFLATE, by default True.
    dstSRS : str, optional
        Reproject the output to `dstSRS`, by default "epsg:4326".
    """
    options_dict = {"dstSRS": dstSRS}
    
    if not isinstance(bb, type(None)):
        options_dict["outputBounds"] = (bb[1][0], bb[0][0], 
                                        bb[1][1], bb[0][1])

    if compress:
        options_dict["creationOptions"] = ["COMPRESS=DEFLATE", "ZLEVEL=8"]

    if isinstance(dest_fp, type(None)):
        log.debug(f"Overwriting input file ({source_fp}).")
        dest_fp = source_fp

    options = gdal.WarpOptions(**options_dict)

    out = gdal.Warp(dest_fp, source_fp, options = options)
    out.FlushCache()
    out = None

    return dest_fp

def apply_mask(a, indices, axis):
    """Select indices in `a` along an axis.

    Parameters
    ----------
    a : np.ndarray
        Array with one dimension more than `indices`.
    indices : np.ndarray
        Array used to index `a`, should have one dimension fewer than `a`.
    axis : int
        Axis along which to select values from `a`.

    Returns
    -------
    np.ndarray
        Array with same shape as `indices`.
    """
    # https://stackoverflow.com/questions/15469302/numpy-3d-to-2d-transformation-based-on-2d-mask-array
    # TODO replace with np.take(....)?

    magic_index = [np.arange(i) for i in indices.shape]
    magic_index = np.ix_(*magic_index)
    magic_index = magic_index[:axis] + (indices,) + magic_index[axis:]
    return a[magic_index]

def reproj_file(file, template, method):
    """Reprojects and resamples a file and return the output as an np.ndarray.

    Parameters
    ----------
    file : {str | gdal.Dataset}
        File to reproject.
    template : {str | gdal.Dataset}
        File to use a reprojection template.
    method : int
        Interpolation method, see 'pywapor.processing_functions.reproject_dataset_example'
        for more info.

    Returns
    -------
    np.ndarray
        The resamples data.
    """
    ds = reproject_dataset_example(file, template, method = method)
    array = open_as_array(ds)
    return array

def combine_dicts(dicts):
    """Combines dictionaries by appending values for identical keys in a list.

    Parameters
    ----------
    dicts : list
        A list of dictionaries.

    Returns
    -------
    dict
        Dictionary with all the values and keys from the dictionaries passed inside 
        the list `dicts`.
    """
    new_dict = dict()
    for d in dicts:
        for key, value in d.items():
            if key in new_dict.keys():
                new_dict[key].append(value)
            else:
                new_dict[key] = [value]
    return new_dict

def get_geoinfo(template_file):
    """Extract relevant geo information from a geotiff file.

    Parameters
    ----------
    template_file : str
        Path to geotiff file.

    Returns
    -------
    tuple
        Tuple with the geotransform, projection, amount of pixels in x-direction and
        amount of pixels in y-direction.
    """
    if isinstance(template_file, str):
        ds = gdal.Open(template_file)
    elif isinstance(template_file, gdal.Dataset):
        ds = template_file
    geo_ex = ds.GetGeoTransform()
    proj_ex = ds.GetProjection()
    size_x_ex = ds.RasterXSize
    size_y_ex = ds.RasterYSize
    return (geo_ex, proj_ex, size_x_ex, size_y_ex)

def Extract_Data_gz(zip_filename, outfilename):
    """Extract a zip-file and removes the zip-file afterwards.

    Parameters
    ----------
    zip_filename : str
        Path to zip-file.
    outfilename : str
        Path to output.
    """
    # TODO maybe move to CHIRPS?
    with gzip.GzipFile(zip_filename, 'rb') as zf:
        file_content = zf.read()
        save_file_content = open(outfilename, 'wb')
        save_file_content.write(file_content)
    save_file_content.close()
    zf.close()
    os.remove(zip_filename)
    
def Save_as_MEM(data='', geo='', projection=''):
    """Create a gdal.Dataset in memory from a np.ndarray.

    Parameters
    ----------
    data : np.ndarray
        Data to put inside the gdal.Dataset.
    geo : list
        Geotransform
    projection : osr.SpatialReference
        Projection to be used in the gdal.Dataset, by default "WGS84".
    """
    # TODO fix default values '' etc.
    driver = gdal.GetDriverByName("MEM")
    dst_ds = driver.Create('', int(data.shape[1]), int(data.shape[0]), 1,
                           gdal.GDT_Float32)
    srse = osr.SpatialReference()
    if projection == '':
        srse.SetWellKnownGeogCS("WGS84")

    else:
        # TODO Remove try/excepts
        try:
            if not srse.SetWellKnownGeogCS(projection) == 6:
                srse.SetWellKnownGeogCS(projection)
            else:
                try:
                    srse.ImportFromEPSG(int(projection))
                except:
                    srse.ImportFromWkt(projection)
        except:
            try:
                srse.ImportFromEPSG(int(projection))
            except:
                srse.ImportFromWkt(projection)
    dst_ds.SetProjection(srse.ExportToWkt())
    dst_ds.GetRasterBand(1).SetNoDataValue(-9999)
    dst_ds.SetGeoTransform(geo)
    dst_ds.GetRasterBand(1).WriteArray(data)
    return(dst_ds)   
    
def Save_as_tiff(name='', data='', geo='', projection=''):
    """Create a geotiff from a np.ndarray.

    Parameters
    ----------
    nme : str
        Path to output file.
    data : np.ndarray
        Data to put inside the geotiff.
    geo : list
        Geotransform
    projection : osr.SpatialReference
        Projection to be used in the geotiff, by default "WGS84".
    """
    # TODO fix default values '' etc.
    # Change no data values
    data[np.isnan(data)] = -9999

    if not os.path.exists(os.path.split(name)[0]):
        os.makedirs(os.path.split(name)[0])
    
    # save as a geotiff
    driver = gdal.GetDriverByName("GTiff")
    dst_ds = driver.Create(name, int(data.shape[1]), int(data.shape[0]), 1,
                           gdal.GDT_Float32, ['COMPRESS=LZW', 'PREDICTOR=3'])
    srse = osr.SpatialReference()
    if projection == '':
        srse.SetWellKnownGeogCS("WGS84")

    # Set the projection, which can be an EPSG code or a well known GeogCS
    else:
        # TODO get rid of all these try/excepts
        try:
            if not srse.SetWellKnownGeogCS(projection) == 6:
                srse.SetWellKnownGeogCS(projection)
            else:
                try:
                    srse.ImportFromEPSG(int(projection))
                except:
                    srse.ImportFromWkt(projection)
        except:
            try:
                srse.ImportFromEPSG(int(projection))
            except:
                srse.ImportFromWkt(projection)
                
    # Save the tiff file
    dst_ds.SetProjection(srse.ExportToWkt())
    dst_ds.GetRasterBand(1).SetNoDataValue(-9999)
    dst_ds.SetGeoTransform(geo)
    dst_ds.GetRasterBand(1).WriteArray(data)
    dst_ds = None
    
    return()
    
def reproject_MODIS(input_name, epsg_to):
    # TODO move this to pywapor.collect.MODIS
    
    # Define the output name
    name_out = ''.join(input_name.split(".")[:-1]) + '_reprojected.tif'
   
    src_ds = gdal.Open(input_name)
    
    # Define target SRS
    dst_srs = osr.SpatialReference()
    dst_srs.ImportFromEPSG(int(epsg_to))
    dst_wkt = dst_srs.ExportToWkt()
    
    error_threshold = 0.125  # error threshold --> use same value as in gdalwarp
    resampling = gdal.GRA_NearestNeighbour
    
    # Call AutoCreateWarpedVRT() to fetch default values for target raster dimensions and geotransform
    tmp_ds = gdal.AutoCreateWarpedVRT( src_ds,
                                   None, # src_wkt : left to default value --> will use the one from source
                                   dst_wkt,
                                   resampling,
                                   error_threshold )
    dst_ds = gdal.GetDriverByName('GTiff').CreateCopy(name_out, tmp_ds)
    dst_ds = None 

    return(name_out)
    
def clip_data(input_file, latlim, lonlim):
    """Clip the extents of a geotiff.

    Parameters
    ----------
    input_file : {str | gdal.Dataset}
        Dataset to be clipped
    latlim : list
        List with lower and upper latitude boundaries.
    lonlim : list
        List with lower and upper longitude boundaries.

    Returns
    -------
    np.ndarray
        The clipped data.
    tuple
        The new geotransform.
    """
    # TODO merge with reproject_clip
    # TODO remove try/except.
    try:
        if input_file.split('.')[-1] == 'tif':
            dest_in = gdal.Open(input_file)
        else:
            dest_in = input_file
    except:
        dest_in = input_file

    # Open Array
    data_in = dest_in.GetRasterBand(1).ReadAsArray()

    # Define the array that must remain
    Geo_in = dest_in.GetGeoTransform()
    Geo_in = list(Geo_in)
    Start_x = np.max([int(np.floor(((lonlim[0]) - Geo_in[0])/ Geo_in[1])),0])
    End_x = np.min([int(np.ceil(((lonlim[1]) - Geo_in[0])/ Geo_in[1])),int(dest_in.RasterXSize)])

    Start_y = np.max([int(np.floor((Geo_in[3] - latlim[1])/ -Geo_in[5])),0])
    End_y = np.min([int(np.ceil(((latlim[0]) - Geo_in[3])/Geo_in[5])), int(dest_in.RasterYSize)])

    #Create new GeoTransform
    Geo_in[0] = Geo_in[0] + Start_x * Geo_in[1]
    Geo_in[3] = Geo_in[3] + Start_y * Geo_in[5]
    Geo_out = tuple(Geo_in)

    data = np.zeros([End_y - Start_y, End_x - Start_x])

    data = data_in[Start_y:End_y,Start_x:End_x]
    dest_in = None

    return(data, Geo_out)    
    
    
def Extract_Data(input_file, output_folder):
    """Extract a zipfile.

    Parameters
    ----------
    input_file : str
        Path to zip-file.
    output_folder : str
        Path to output folder.
    """
    # TODO merge with pywapor.general.processing_functions.Extract_Data_gz
    # extract the data, used in SRTM and Globcover
    z = zipfile.ZipFile(input_file, 'r')
    z.extractall(output_folder)
    z.close()    

def reproj_ds(source_file, example_file, ndv = -9999, override_dtype = True):
    """Reproject and resample (nearest neighbour) a geotiff to match with
    `example_file`.

    Parameters
    ----------
    source_file : str
        Path to source file.
    example_file : str
        Path to example file.
    ndv : int, optional
        No-data-value to be used in the reprojected dataset, by default -9999
    override_dtype : bool, optional
        Use the same dtype as `example_file` (True) or as `source_file` (False), 
        by default True.

    Returns
    -------
    gdal.Dataset
        The reprojected dataset.
    """
    # TODO merge with reproject_dataset_example
    source_ds = gdal.Open(source_file)

    example_ds = gdal.Open(example_file)
    example_xsize = example_ds.RasterXSize
    example_ysize = example_ds.RasterYSize

    if override_dtype:
        example_dtype = example_ds.GetRasterBand(1).DataType
    else:
        example_dtype = source_ds.GetRasterBand(1).DataType

    mem_drv = gdal.GetDriverByName('MEM')
    dest_ds = mem_drv.Create('', example_xsize, example_ysize, 1, example_dtype)
    dest_ds.SetGeoTransform(example_ds.GetGeoTransform())
    dest_ds.SetProjection(example_ds.GetProjection())
    dest_ds.GetRasterBand(1).Fill(ndv)
    dest_ds.GetRasterBand(1).SetNoDataValue(ndv)

    gdal.ReprojectImage(source_ds,
                        dest_ds,
                        source_ds.GetProjection(),
                        dest_ds.GetProjection(),
                        gdal.GRA_NearestNeighbour)

    return dest_ds

def reproject_dataset_example(dataset, dataset_example, method=1):
    """Reproject and resample a dataset to match with `dataset_example`.

    Parameters
    ----------
    dataset : {str | gdal.Dataset}
        The dataset to be reprojected.
    dataset_example : {str | gdal.Dataset}
        The dataset to be used as a reprojection example.
    method : int, optional
        Select which method to use for resampling, with:
        1 = gdal.GRA_NearestNeighbour,
        2 = gdal.GRA_Bilinear,
        3 = gdal.GRA_Lanczos,
        4 = gdal.GRA_Average,
        5 = gdal.GRA_Cubic,
        6 = gdal.GRA_CubicSpline,
        7 = gdal.GRA_Mode,
        8 = gdal.GRA_Max,
        9 = gdal.GRA_Min,
        10 = gdal.GRA_Med,
        11 = gdal.GRA_Q1,
        12 = gdal.GRA_Q3,
        13 = gdal.GRA_Sum, by default 1.

    Returns
    -------
    gdal.Dataset
        The reprojected dataset.

    """
    # TODO merge with reproj_ds
    # open dataset that must be transformed
    if isinstance(dataset, str):
        g = gdal.Open(dataset)
    elif isinstance(dataset, gdal.Dataset):
        g = dataset
    else:
        raise ValueError
    epsg_from = Get_epsg(g)

    #exceptions
    if epsg_from == 9001:
        epsg_from = 5070

    # open dataset that is used for transforming the dataset
    if isinstance(dataset_example, str):
        gland = gdal.Open(dataset_example)
    elif isinstance(dataset_example, gdal.Dataset):
        gland = dataset_example
    else:
        raise ValueError
    epsg_to = Get_epsg(gland)

    # Set the EPSG codes
    osng = osr.SpatialReference()
    osng.ImportFromEPSG(epsg_to)
    wgs84 = osr.SpatialReference()
    wgs84.ImportFromEPSG(epsg_from)

    # Get shape and geo transform from example
    geo_land = gland.GetGeoTransform()
    col=gland.RasterXSize
    rows=gland.RasterYSize

    # Create new raster
    mem_drv = gdal.GetDriverByName('MEM')
    dest = mem_drv.Create('', col, rows, 1, gdal.GDT_Float32)
    dest.SetGeoTransform(geo_land)
    dest.SetProjection(osng.ExportToWkt())

    # https://gis.stackexchange.com/questions/158503/9999-no-data-value-becomes-0-when-writing-array-to-gdal-memory-file
    ndv = g.GetRasterBand(1).GetNoDataValue()
    band = dest.GetRasterBand(1)
    band.SetNoDataValue(ndv)
    band.Fill(ndv, 0.0)

    # Perform the projection/resampling
    if method == 1:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_NearestNeighbour)
    if method == 2:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Bilinear)
    if method == 3:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Lanczos)
    if method == 4:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Average)
    if method == 5:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Cubic)
    if method == 6:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_CubicSpline)
    if method == 7:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Mode)
    if method == 8:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Max)
    if method == 9:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Min)
    if method == 10:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Med)
    if method == 11:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Q1)
    if method == 12:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Q3)
    if method == 13:
        gdal.ReprojectImage(g, dest, wgs84.ExportToWkt(), osng.ExportToWkt(), gdal.GRA_Sum)
        
    return(dest)

def Get_epsg(g):
    # TODO deprecate
    try:
        # Get info of the dataset that is used for transforming
        g_proj = g.GetProjection()
        Projection=g_proj.split('EPSG","')
        epsg_to=int((str(Projection[-1]).split(']')[0])[0:-1])            
    except:
        epsg_to=4326

    return(epsg_to)

def open_as_array(input):
    """Open dataset as array, replacing no-data-values with np.nan.

    Parameters
    ----------
    input : {str | gdal.Dataset}
        The dataset to open.

    Returns
    -------
    np.ndarray
        The data contained inside `input`.
    """
    if isinstance(input, str):
        ds = gdal.Open(input)
    elif isinstance(input, gdal.Dataset):
        ds = input
    array = ds.GetRasterBand(1).ReadAsArray().astype(float)
    ndv = ds.GetRasterBand(1).GetNoDataValue()
    array[array == ndv] = np.nan
    return array  

def Open_tiff_array(filename='', band=''):
    # TODO deprecate
    f = gdal.Open(filename)
    if f is None:
        print('%s does not exists' %filename)
    else:
        if band == '':
            band = 1
        Data = f.GetRasterBand(band).ReadAsArray()
    return(Data)

def Open_array_info(filename=''):
    # TODO deprecate
    try:
        if filename.split('.')[-1] == 'tif':
            f = gdal.Open(r"%s" %filename)
        else:
            f = filename
    except:
            f = filename       
    try:
        geo_out = f.GetGeoTransform()
        proj = f.GetProjection()
        size_X = f.RasterXSize
        size_Y = f.RasterYSize
        f = None
    except:
        print('%s does not exists' %filename)
        
    return(geo_out, proj, size_X, size_Y)

def gap_filling(dataset, NoDataValue, method = 1):
    # TODO deprecate
    try:
        if dataset.split('.')[-1] == 'tif':
            # Open the numpy array
            data = Open_tiff_array(dataset)
            Save_as_tiff_bool = 1
        else:
            data = dataset
            Save_as_tiff_bool = 0
    except:
        data = dataset
        Save_as_tiff_bool = 0

    # fill the no data values
    if NoDataValue == np.nan:
        mask = ~(np.isnan(data))
    else:
        mask = ~(data==NoDataValue)
    xx, yy = np.meshgrid(np.arange(data.shape[1]), np.arange(data.shape[0]))
    xym = np.vstack( (np.ravel(xx[mask]), np.ravel(yy[mask])) ).T
    data0 = np.ravel( data[:,:][mask] )

    if method == 1:
        interp0 = NearestNDInterpolator( xym, data0 )
        data_end = interp0(np.ravel(xx), np.ravel(yy)).reshape( xx.shape )

    if method == 2:
        interp0 = LinearNDInterpolator( xym, data0 )
        data_end = interp0(np.ravel(xx), np.ravel(yy)).reshape( xx.shape )

    if Save_as_tiff_bool == 1:
        EndProduct=dataset[:-4] + '_GF.tif'

        # collect the geoinformation
        geo_out, proj, size_X, size_Y = Open_array_info(dataset)

        # Save the filled array as geotiff
        Save_as_tiff(name=EndProduct, data=data_end, geo=geo_out, projection=proj)

    else:
        EndProduct = data_end

    return (EndProduct)  
    
def calc_dlat_dlon(geo_out, size_X, size_Y, lat_lon = None):
    """
    Calculated the dimensions of each pixel in meter.

    Parameters
    ----------
    geo_out: list
        Geotransform function of the array.
    size_X: int
        Number of pixels in x-direction.
    size_Y: int
        Number of pixels in y-direction.

    Returns
    -------
    np.ndarray
        Size of every pixel in the y-direction in meters.
    dlon: array
        Size of every pixel in the x-direction in meters.
    """
    if isinstance(lat_lon, type(None)):
        # Create the lat/lon rasters
        lon = np.arange(size_X + 1)*geo_out[1]+geo_out[0] - 0.5 * geo_out[1]
        lat = np.arange(size_Y + 1)*geo_out[5]+geo_out[3] - 0.5 * geo_out[5]
    else:
        lat, lon = lat_lon

    dlat_2d = np.array([lat,]*int(np.size(lon,0))).transpose()
    dlon_2d =  np.array([lon,]*int(np.size(lat,0)))

    # Radius of the earth in meters
    R_earth = 6371000

    # Calculate the lat and lon in radians
    lonRad = dlon_2d * np.pi/180
    latRad = dlat_2d * np.pi/180

    # Calculate the difference in lat and lon
    lonRad_dif = abs(lonRad[:,1:] - lonRad[:,:-1])
    latRad_dif = abs(latRad[:-1] - latRad[1:])

    # Calculate the distance between the upper and lower pixel edge
    a = np.sin(latRad_dif[:,:-1]/2) * np.sin(latRad_dif[:,:-1]/2)
    clat = 2 * np.arctan2(np.sqrt(a), np.sqrt(1-a))
    dlat = R_earth * clat

    # Calculate the distance between the eastern and western pixel edge
    b = np.cos(latRad[1:,:-1]) * np.cos(latRad[:-1,:-1]) * np.sin(lonRad_dif[:-1,:]/2) * np.sin(lonRad_dif[:-1,:]/2)
    clon = 2 * np.arctan2(np.sqrt(b), np.sqrt(1-b))
    dlon = R_earth * clon

    return(dlat, dlon)


# -*- coding: utf-8 -*-
"""
Generates input data for `pywapor.et_look`.
"""
import os
import glob
import warnings
import shutil
import pywapor
import xarray as xr
import pandas as pd
import numpy as np
import pywapor.general as g
import pywapor.general.processing_functions as pf
import pywapor.general.pre_defaults as defaults
from datetime import datetime as dat
from pywapor.collect.downloader import collect_sources
from pywapor.general.logger import log, adjust_logger
from pywapor.enhancers.apply_enhancers import apply_enhancer
from pywapor.general.compositer import calculate_ds
import copy

def main(project_folder: str, 
        startdate: str, 
        enddate: str, 
        latlim: list,
        lonlim: list, 
        level = "level_1", 
        composite_length = "DEKAD", 
        extra_composite_enhancements = {},
        extra_source_enhancements = {}, 
        extra_source_locations = {}, 
        diagnostics = {}, 
        se_root_version = "v2", 
        process_vars = "all"):
    """Generates an input file for `pywapor.et_look` based on, among others, a
    bounding-box and time period.

    Parameters
    ----------
    project_folder : str
        Path to folder in which (intermediate) data will be stored.
    startdate : str
        Start of the period for which to generate the input data, formatted as
        "YYYY-MM-DD".
    enddate : str
        End of the period for which to generate the input data, formatted as
        "YYYY-MM-DD".
    latlim : List[float, float]
        Latitude limits of the bounding-box, first value should be smaller than
        second value.
    lonlim : List[float, float]
        Longitude limits of the bounding-box, first value should be smaller than
        second value.
    level : "level_1" | "level_2" | dict, optional
        Controls which sources are collected for the various variables, 
        by default "level_1".
    composite_length : "DEKAD" | int, optional
        Length of the composites in days in case an integer is given, by default "DEKAD".
    extra_composite_enhancements : dict, optional
        Functions to be applied to specified variables after the composites have been calculated, by default {}.
    extra_source_enhancements : dict, optional
        Functions to be applied to specified variables before the composites have been calculated, by default {}.
    extra_source_locations : dict, optional
        Paths in which to look for sideloaded data, by default {}.
    diagnostics : dict, optional
        Points for which to generate plots, by default {}.
    se_root_version : "v2" | "dev", optional
        Which version of pywapor.se_root to use to generate the se_root variable, by default "v2".
    process_vars : "all" | list, optional
        Which variables to process, by default "all", which is equivalent to providing a list with
        `ndvi`, `p_24`, `se_root`, `r0`, `z`, `lulc`, `ra_24`, `t_air_24`, 
        `t_air_min_24`, `t_air_max_24`, `u2m_24`, `v2m_24`, `p_air_0_24`, `qv_24`,
        `lw_offset`, `lw_slope`, `r0_bare`, `r0_full`, `rn_offset`, 
        `rn_slope`, `t_amp_year`, `t_opt`, `vpd_slope` and `z_oro`.

    Returns
    -------
    xr.Dataset
        Dataset with all the required data for pywapor.et_look.
    str
        Path to the netCDF file containing the generated dataset.

    Examples
    --------
    >>> import pywapor
    >>> project_folder = r"path/to/my_project_folder/"
    >>> startdate = "2021-07-01"
    >>> enddate = "2021-07-11"
    >>> latlim = [28.9, 29.7]
    >>> lonlim = [30.2, 31.2]
    >>> composite_length = 8
    >>> my_custom_level = {"ndvi": ["LS7"],
    ...                    "t_air_24": ["GEOS5"],
    ...                    "level_name": "my_custom_name"}
    >>> extra_sources = {("LS7", "ndvi"): "path/to/my/landsat/folder"}
    >>> source_enhancers = {("LS7", "ndvi"): [pywapor.general.enhancers.gap_fill.gap_fill]}
    >>> composite_enhancers = {("t_air_24"): []}
    >>> ds, fh = pywapor.pre_et_look.main(project_folder, startdate, enddate, 
    ...                             latlim, lonlim, level = my_custom_level, 
    ...                             extra_source_locations = extra_sources,
    ...                             composite_length = composite_length,
    ...                             extra_source_enhancements = source_enhancers,
    ...                             extra_composite_enhancements = composite_enhancers,
    ...                             process_vars = ["ndvi", "t_air_24"])
    >>> print(fh)
    path/to/my_project_folder/my_custom_name/et_look_input.nc
    """

    # Create project folder.
    if not os.path.exists(project_folder):
        os.makedirs(project_folder)

    # Initiate logger.
    log_write = True
    log_level = "INFO"
    adjust_logger(log_write, project_folder, log_level)

    log.info("> PRE_ET_LOOK").add()

    # Disable Warnings
    warnings.filterwarnings('ignore')

    # Calculate relevant dates.
    sdate = dat.strptime(startdate, "%Y-%m-%d").date()
    edate = dat.strptime(enddate, "%Y-%m-%d").date()
    epochs_info = create_dates(sdate, edate, composite_length)

    # Load required variable sources.
    if isinstance(level, str):
        levels = g.variables.get_source_level_selections()
        source_selection = levels[level]
        level_name = level
    elif isinstance(level, dict):
        source_selection = copy.copy(level)
        if "level_name" in source_selection.keys():
            level_name = source_selection.pop("level_name")
        else:
            level_name = "custom"

    # Define folders.
    level_folder, temp_folder, raw_folder = get_folders(project_folder, level_name)
    if diagnostics:
        diagnostics["folder"] = os.path.join(level_folder, "graphs")

    # Define download arguments.
    dl_args = {
                "Dir": raw_folder, 
                "latlim": latlim,
                "lonlim": lonlim,
                "Startdate": startdate, 
                "Enddate": enddate,
                }

    # List of variables to process. Resampling resolution is taken from 
    # first item in the list.
    if process_vars == "all":
        all_vars = [
                    "ndvi",
                    "p_24",
                    "se_root",
                    "r0", 
                    "z",
                    "lulc",
                    "ra_24",
                    't_air_24', 
                    't_air_min_24', 't_air_max_24', 
                    'u2m_24', 'v2m_24', 'p_air_0_24', 'qv_24',
                    'lw_offset', 'lw_slope', 'r0_bare', 'r0_full', 'rn_offset', 
                    'rn_slope', 't_amp_year', 't_opt', 'vpd_slope', 'z_oro',
                    # # 'land_mask', 'rs_min', 'z_obst_max', # these are generated from lulc
                    ]
    else:
        all_vars = process_vars

    datasets = list()

    # Loop over the variables, resampling, enhancing and compositing each one.
    for var in all_vars:

        log.info(f"# {var}")

        # Run pre_se_root and se_root.
        if var == "se_root":
            ds_in = pywapor.pre_se_root.main(project_folder, startdate, enddate, latlim, 
                                            lonlim, level = level,
                                            extra_source_locations = extra_source_locations)[0]
            raw_files = [pywapor.se_root.main(ds_in, se_root_version = se_root_version, 
                                                export_to_tif = True, export_vars = "default")["se_root"]]
        # Download RAW-data.
        else:
            raw_files = collect_sources(var, source_selection[var], 
                                        dl_args, extra_source_locations)

        # Define resampling parameters.
        if 'example_info' not in vars():
            example_info = pf.select_template(raw_files)
            # example_info[1].to_netcdf(r"/Users/hmcoerver/pywapor_notebooks/example_ds.nc")

        # Create composites.
        ds = unraw_all(var, raw_files, epochs_info, temp_folder, 
                        example_info[1], diagnostics = diagnostics, 
                        extra_source_enhancements = extra_source_enhancements)

        # Store variable composites in a list.
        datasets.append(ds)

    # Merge all the variable composites into one xr.Dataset.
    ds = xr.merge(datasets, combine_attrs = "drop_conflicts")
    ds.attrs = {}

    log.info("> Composite enhancers.").add()

    # Rename variables.
    ds = ds.rename_vars({x: x.replace("_composite", "") for x in list(ds.variables)})

    # Define composite enhancements.
    composite_enhancements = defaults.composite_enhancements_defaults()
    composite_enhancements = {**composite_enhancements, **extra_composite_enhancements}

    # Apply composite enhancements.
    for variable, enhancers in composite_enhancements.items():
        if variable not in list(ds.keys()):
            continue
        log.info(f"# {variable}")
        for enhancer in enhancers:
            ds, label = apply_enhancer(ds, variable, enhancer)
            ds, _ = calculate_ds(ds, label = label)

    log.sub().info("< Composite enhancers.")

    ds = g.variables.fill_attrs(ds)

    ds.attrs["geotransform"] = example_info[2][0]
    ds.attrs["projection"] = example_info[2][1]
    ds.attrs["pixel_size"] = example_info[3]
    ds.attrs["example_file"] = example_info[0]

    if "spatial_ref" in list(ds.coords):
        ds = ds.drop_vars("spatial_ref")

    # Add constants to ds
    ds = ds.assign(defaults.constants_defaults())

    # Save pre_et_look-output (i.e. et_look-input).
    all_vars = [var for var in list(ds.variables) if "lon" in ds[var].coords 
                                                and "lat" in ds[var].coords]
    out_fh = os.path.join(level_folder, "et_look_input.nc")
    ds, out_fh = calculate_ds(ds, out_fh, label = "--> Saving results.", 
                                encoding = {k: {"dtype": "float32"} for k in all_vars}
                                )

    # Remove temporary files.
    for fh in glob.glob(os.path.join(temp_folder, "*.nc")):
        os.remove(fh)
    if len(os.listdir(temp_folder)) == 0:
        shutil.rmtree(temp_folder)

    # Reset working directory.
    os.chdir(project_folder)

    log.sub().info("< PRE_ET_LOOK")

    return ds, out_fh

def get_folders(project_folder, level_name = None):
    """Define some folders based on a root-folder.

    Parameters
    ----------
    project_folder : str
        Path to root-folder.
    level_name : str, optional
        Name of the source selection level, by default None.

    Returns
    -------
    str
        Path to level-folder.
    str
        Path to temporary-data-folder.
    str
        Path to raw-folder.
    """
    if not isinstance(level_name, type(None)):
        level_folder = os.path.join(project_folder, level_name)
    else:
        level_folder = None
    temp_folder = os.path.join(project_folder, "temporary")
    raw_folder = os.path.join(project_folder, "RAW")
    return level_folder, temp_folder, raw_folder

def unraw_all(var, raw_files, epochs_info, temp_folder, 
                example_ds, diagnostics = {}, extra_source_enhancements = {}):
    """Converts raw downloaded files into composites.

    Parameters
    ----------
    var : str
        Name of the variable to process.
    raw_files : list
        List of lists containing the paths to the geoTIFF files for which to compute the
        composites. All paths in one sublist refer to one source.
    epochs_info : tuple
        Contains three lists with the index-number, starttime and endtime of each
        epoch.
    temp_folder : str
        Path to folder in which to store temporary data.
    example_ds : xr.Dataset
        Dataset used to interpolate data (spatially).
    diagnostics : dict, optional
        Points for which to generate plots, by default {}.
    extra_source_enhancements : dict, optional
        Functions to be applied to specified variables before the composites have been calculated, by default {}.

    Returns
    -------
    xr.Dataset
        Dataset with the composite data for the respective variable.
    """

    cmeta = defaults.composite_defaults()[var]

    ds = g.compositer.main(cmeta, raw_files, epochs_info, temp_folder, example_ds,
                            extra_source_enhancements = extra_source_enhancements)
    if diagnostics:
        _ = g.compositer.main(cmeta, raw_files, epochs_info, None, example_ds, 
                                    lean_output = False, diagnostics = diagnostics, 
                                    extra_source_enhancements = extra_source_enhancements)

    return ds

def create_dates(sdate, edate, period_length):
    """Generates the variable `epoch_info` used by `pywapor.unraw_all`.

    Parameters
    ----------
    sdate : str or datetime.datetime
        Date from which to start.
    edate : str or datetime.datetime
        Enddate.
    period_length : int or str
        Amount of days between the returned dates, can also be "DEKAD" in which
        case it will split each month in roughly 10-day segments.

    Returns
    -------
    np.ndarray
        Index number of the epoch.
    np.ndarray
        Start-datetimes of the epochs.
    np.ndarray
        End-datetimes of the epochs.
    """
    if isinstance(sdate, str):
        sdate = dat.strptime(sdate, "%Y-%m-%d")
    if isinstance(edate, str):
        edate = dat.strptime(edate, "%Y-%m-%d")
    if period_length == "DEKAD":
        dekad1 = pd.date_range(sdate - pd.to_timedelta(sdate.day - 1, unit='d'), edate, freq = "MS")
        dekad2 = dekad1 + pd.Timedelta(10, unit='D')
        dekad3 = dekad1 + pd.Timedelta(20, unit='D')
        dates = np.sort(np.array([dekad1, dekad2, dekad3]).flatten())
        no_periods = dates.size - 1
    else:
        days = (edate - sdate).days
        no_periods = int(days // period_length + 1)
        dates = pd.date_range(sdate, periods = no_periods + 1 , freq = f"{period_length}D")
    periods_start = dates[:-1]
    mask = np.all([periods_start < np.datetime64(edate), 
                   periods_start >= np.datetime64(sdate)], axis = 0)
    epochs = np.arange(0, no_periods, 1)[mask]
    periods_end = dates[1:][mask]
    periods_start = periods_start[mask]
    return epochs, periods_start, periods_end
         
if __name__ == "__main__":

    project_folder = r"/Users/hmcoerver/pywapor_notebooks"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    startdate = "2021-07-01"
    enddate = "2021-07-11"
    # startdate = "2021-07-06"
    # enddate = "2021-07-07"
    composite_length = "DEKAD"
    # composite_length = 1
    # level = "level_1"
    extra_sources = None
    extra_source_locations = None
    se_root_version = "v2"

    diagnostics = { # label          # lat      # lon
                    "water":	    (29.44977,	30.58215),
                    "desert":	    (29.12343,	30.51222),
                    "agriculture":	(29.32301,	30.77599),
                    "urban":	    (29.30962,	30.84109),
                    }
    # diagnostics = None

    # level = {
    #     # Main inputs
    #     "ndvi":         ["LS7"],
    #     "r0":           ["LS7"],
    #     "lst":          ["LS7"],
    #     "lulc":         ["WAPOR"],
    #     "z":            ["SRTM"],
    #     "p_24":         ["CHIRPS"],
    #     "ra_24":        ["MERRA2"],

    #     # Daily meteo 
    #     't_air_24':     ["GEOS5"],
    #     't_air_min_24': ["GEOS5"], 
    #     't_air_max_24': ["GEOS5"],
    #     'u2m_24':       ["GEOS5"],
    #     'v2m_24':       ["GEOS5"],
    #     'p_air_0_24':   ["GEOS5"],
    #     'qv_24':        ["GEOS5"],

    #     # Instanteneous meteo
    #     "t_air_i":      ["GEOS5"],
    #     "u2m_i":        ["GEOS5"],
    #     "v2m_i":        ["GEOS5"],
    #     "qv_i":         ["GEOS5"],
    #     "wv_i":         ["GEOS5"],
    #     "p_air_i":      ["GEOS5"],
    #     "p_air_0_i":    ["GEOS5"],

    #     # Temporal constants
    #     "lw_offset":    ["STATICS"],
    #     "lw_slope":     ["STATICS"],
    #     "r0_bare":      ["STATICS"],
    #     "r0_full":      ["STATICS"],
    #     "rn_offset":    ["STATICS"],
    #     "rn_slope":     ["STATICS"],
    #     "t_amp_year":   ["STATICS"],
    #     "t_opt":        ["STATICS"],
    #     "vpd_slope":    ["STATICS"],
    #     "z_oro":        ["STATICS"],

    #     # Level name
    #     "level_name": "sideloading",
    # }

    # Give product folder.
    # extra_source_locations = {
    #     ("LS7", "ndvi"): r"/Users/hmcoerver/pywapor_notebooks/my_landsat_folder/NDVI",
    #     ("LS7", "lst"): r"/Users/hmcoerver/pywapor_notebooks/my_landsat_folder/LST",
    #     ("LS7", "r0"): r"/Users/hmcoerver/pywapor_notebooks/my_landsat_folder/ALBEDO",
    # }

    ds_in, fh_in = main(project_folder, startdate, enddate, latlim, lonlim, 
        diagnostics = diagnostics, composite_length = composite_length,
        extra_source_locations = extra_source_locations, se_root_version = se_root_version)

    # fh_in = r"/Users/hmcoerver/pywapor_notebooks/level_1/et_look_input___.nc"

    # out = pywapor.et_look.main(fh_in, export_vars = "all")

    # param = "t_air_24"
    # sources = ["GEOS5"]

    # dl_args = {
    #             "Dir": os.path.join(project_folder, "RAW"), 
    #             "latlim": latlim, 
    #             "lonlim": lonlim, 
    #             "Startdate": startdate, 
    #             "Enddate": enddate,
    #             }

    # files = collect_sources(param, sources, dl_args)






# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/CHIRPS
"""
import os
from osgeo import gdal
import numpy as np
import pandas as pd
from ftplib import FTP
import tqdm

def DownloadData(Dir, Startdate, Enddate, latlim, lonlim, Waitbar):
    """
    This function downloads CHIRPS daily or monthly data

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax] (values must be between -50 and 50)
    lonlim -- [xmin, xmax] (values must be between -180 and 180)
    Waitbar -- 1 (Default) will print a waitbar
    cores -- The number of cores used to run the routine. It can be 'False'
             to avoid using parallel computing routines.
    TimeCase -- String equal to 'daily' or 'monthly'
    """
	
    # Define timestep for the timedates
    output_folder = os.path.join(Dir, "CHIRPS", "Precipitation")

    # make directory if it not exists
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)

	# check time variables
    if not Startdate:
        Startdate = pd.Timestamp('1981-01-01')
    if not Enddate:
        Enddate = pd.Timestamp('Now')

    # Create days
    Dates = pd.date_range(Startdate, Enddate, freq='D')

    if Waitbar:
        waitbar = tqdm.tqdm(desc= f"Tile: 0 / {int(len(Dates))}",
                            position = 0,
                            # total=total_size,
                            unit='Bytes',
                            unit_scale=True,)
    else:
        waitbar = None

    # Check space variables
    if latlim[0] < -50 or latlim[1] > 50:
        print('Latitude above 50N or below 50S is not possible.'
               ' Value set to maximum')
        latlim[0] = np.max(latlim[0], -50)
        latlim[1] = np.min(lonlim[1], 50)
    if lonlim[0] < -180 or lonlim[1] > 180:
        print('Longitude must be between 180E and 180W.'
               ' Now value is set to maximum')
        lonlim[0] = np.max(latlim[0], -180)
        lonlim[1] = np.min(lonlim[1], 180)

    # Define IDs
    yID = 2000 - np.int16(np.array([np.ceil((latlim[1] + 50)*20),
                                    np.floor((latlim[0] + 50)*20)]))
    xID = np.int16(np.array([np.floor((lonlim[0] + 180)*20),
                             np.ceil((lonlim[1] + 180)*20)]))

    # Pass variables to parallel function and run
    args = [output_folder, xID, yID, lonlim, latlim]
    for Date in Dates:
        RetrieveData(Date, args, waitbar)

    results = True
    
    # remove raw files
    import glob
    os.chdir(output_folder)
    try:
        files_raw = glob.glob("chirps-v2.0*.tif")
        for file_raw in files_raw:
            outfilename = os.path.join(output_folder, file_raw)
            # delete old tif file
            os.remove(outfilename)
    except:
        pass        
        
    return results

def RetrieveData(Date, args, waitbar):
    """
    This function retrieves CHIRPS data for a given date from the
    ftp://chg-ftpout.geog.ucsb.edu server.

    Keyword arguments:
    Date -- 'yyyy-mm-dd'
    args -- A list of parameters defined in the DownloadData function.
    """
    
    # WA+ modules
    import pywapor.general.processing_functions as PF  

    # Argument
    [output_folder, xID, yID, lonlim, latlim] = args

    # Define output
    DirFileEnd = os.path.join(output_folder,'P_CHIRPS.v2.0_mm-day-1_daily_%s.%02s.%02s.tif' %(Date.strftime('%Y'), Date.strftime('%m'), Date.strftime('%d')))

    if not os.path.exists(DirFileEnd):

        # open ftp server
        ftp = FTP("ftp.chc.ucsb.edu", "", "")
        ftp.login()
    
    	# Define FTP path to directory
        pathFTP = 'pub/org/chg/products/CHIRPS-2.0/global_daily/tifs/p05/%s/' %Date.strftime('%Y')
    
        # find the document name in this directory
        ftp.cwd(pathFTP)
        listing = []
    
    	# read all the file names in the directory
        ftp.retrlines("LIST", listing.append)
    
    	# create all the input name (filename) and output (outfilename, filetif, DiFileEnd) names
        filename = 'chirps-v2.0.%s.%02s.%02s.tif.gz' %(Date.strftime('%Y'), Date.strftime('%m'), Date.strftime('%d'))
        outfilename = os.path.join(output_folder,'chirps-v2.0.%s.%02s.%02s.tif' %(Date.strftime('%Y'), Date.strftime('%m'), Date.strftime('%d')))

        # download the global rainfall file

        local_filename = os.path.join(output_folder, filename)
        total_size = ftp.size(filename)
        if not isinstance(waitbar, type(None)):
            waitbar_i = int(waitbar.desc.split(" ")[1])
            waitbar_desc = str(waitbar.desc)
            waitbar.set_description_str(waitbar_desc.replace(f": {waitbar_i} /", f": {waitbar_i+1} /"))
            waitbar.reset(total = total_size)
        with open(local_filename, "wb") as lf:
            def callback(data):
                lf.write(data)
                if not isinstance(waitbar, type(None)):
                    waitbar.update(len(data))
            ftp.retrbinary("RETR " + filename, callback, 8192) 
        PF.Extract_Data_gz(local_filename, outfilename)

        # open tiff file
        dest = gdal.Open(outfilename)
        dataset = dest.GetRasterBand(1).ReadAsArray()

        # clip dataset to the given extent
        data = dataset[yID[0]:yID[1], xID[0]:xID[1]]
        data[data < 0] = -9999

        # save dataset as geotiff file
        geo = [lonlim[0], 0.05, 0, latlim[1], 0, -0.05]
        
        PF.Save_as_tiff(name=DirFileEnd, data=data, geo=geo, projection="WGS84")
    else:
        if not isinstance(waitbar, type(None)):
            waitbar_i = int(waitbar.desc.split(" ")[1])
            waitbar_desc = str(waitbar.desc)
            waitbar.set_description_str(waitbar_desc.replace(f": {waitbar_i} /", f": {waitbar_i+1} /"))

    return True

if __name__ == "__main__":

    import pywapor

    Dir =r"/Volumes/Data/FAO/temp_test"
    Startdate = "2021-06-01"
    Enddate = "2021-07-01"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]

    DownloadData(Dir, Startdate, Enddate, latlim, lonlim, Waitbar = 1)



import numpy as np
import pywapor.general.processing_functions as pf
import xarray as xr

def to_slope(ds, var, out_var = None):
    """Calculate the gradient of `var` (which would usually be `z` or elevation).

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing var.
    var : str
        Variable name.
    out_var : str, optional
        New variable name in case `var` should not be overwritten, by default None.

    Returns
    -------
    xr.Dataset
        Enhanced dataset.
    """
    dlat, dlon = pf.calc_dlat_dlon(None, None, None, (ds.lat.values, ds.lon.values))            

    dem = ds[var].values

    x, y = np.gradient(dem, np.nanmean(dlon), np.nanmean(dlat))
    hypotenuse_array = np.hypot(x,y)
    slope = np.arctan(hypotenuse_array) * 180.0 / np.pi

    new_data = xr.DataArray(slope, 
                                    coords = {
                                                "lat": ds.lat, 
                                                "lon": ds.lon})
    
    new_data.attrs = {"sources": ["dem.to_slope()"]}

    if not isinstance(out_var, type(None)):
        ds[out_var] = new_data
    else: 
        ds[var] = new_data

    return ds

def to_aspect(ds, var, out_var = None):
    """Calculate the aspect of `var` (which would usually be `z` or elevation).

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing var.
    var : str
        Variable name.
    out_var : str, optional
        New variable name in case `var` should not be overwritten, by default None.

    Returns
    -------
    xr.Dataset
        Enhanced dataset.
    """
    dlat, dlon = pf.calc_dlat_dlon(None, None, None, (ds.lat.values, ds.lon.values))            

    dem = ds[var].values

    x, y = np.gradient(dem, np.nanmean(dlon), np.nanmean(dlat))

    aspect = np.arctan2(y/np.nanmean(dlat), 
                        -x/np.nanmean(dlon)) * 180.0 / np.pi + 180.0
    
    new_data = xr.DataArray(aspect, 
                                    coords = {
                                                "lat": ds.lat, 
                                                "lon": ds.lon})
    
    new_data.attrs = {"sources": ["dem.to_aspect()"]}

    if not isinstance(out_var, type(None)):
        ds[out_var] = new_data
    else: 
        ds[var] = new_data

    return ds

def to_lat(ds, var, out_var = None):
    """Calculate the latitude in degrees of every pixel.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing var.
    var : str
        Variable name.
    out_var : str, optional
        New variable name in case `var` should not be overwritten, by default None.

    Returns
    -------
    xr.Dataset
        Enhanced dataset.
    """
    lat_deg = np.rot90(np.tile(ds.lat, ds.lon.size).reshape((ds.lon.size,ds.lat.size)), 3)

    new_data = xr.DataArray(lat_deg, 
                                    coords = {
                                                "lat": ds.lat, 
                                                "lon": ds.lon})
    
    new_data.attrs = {"sources": ["dem.to_lat()"]}

    if not isinstance(out_var, type(None)):
        ds[out_var] = new_data
    else: 
        ds[var] = new_data
    
    return ds

def to_lon(ds, var, out_var = None):
    """Calculate the longitude in degrees of every pixel.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing var.
    var : str
        Variable name.
    out_var : str, optional
        New variable name in case `var` should not be overwritten, by default None.

    Returns
    -------
    xr.Dataset
        Enhanced dataset.
    """
    lon_deg = np.tile(ds.lon, ds.lat.size).reshape((ds.lat.size,ds.lon.size))

    new_data = xr.DataArray(lon_deg, 
                                    coords = {
                                                "lat": ds.lat, 
                                                "lon": ds.lon})
    
    new_data.attrs = {"sources": ["dem.to_lon()"]}

    if not isinstance(out_var, type(None)):
        ds[out_var] = new_data
    else: 
        ds[var] = new_data

    return ds


# -*- coding: utf-8 -*-
"""
This WAPOR toolbox is a set of functions to collect and run the WAPOR ET model
"""

from pywapor import general, et_look_dev, et_look_v2, pre_et_look, et_look, collect, post_et_look, pre_se_root, se_root, enhancers

__all__ = ['general', 'et_look_dev', 'et_look_v2', 'pre_et_look', 'et_look', 'collect', 'post_et_look', 'pre_se_root', 'se_root', 'enhancers']

__version__ = '2.4.2'


# -*- coding: utf-8 -*-
"""
Code to run the SERoot model.
"""
import xarray as xr
import os
import numpy as np
import pandas as pd
import pywapor
import pywapor.general as g
import pywapor.general.processing_functions as PF
import pywapor.et_look_dev as ETLook_dev
import pywapor.et_look_v2 as ETLook_v2
from pywapor.general.logger import log
from pywapor.general.compositer import calculate_ds
import copy

def main(input_data, se_root_version = "v2", export_vars = "default", export_to_tif = False):
    """Runs the ETLook model over the provided input data.

    Parameters
    ----------
    input_data : str | xr.Dataset
        Dataset generated by `pywapor.pre_se_root`.
    se_root_version : "v2" | "dev", optional
        Which version of the SERoot model to use, by default "v2".
    export_vars : "default" | "all" | list, optional
        Specify which variables to save inside the output file. `"Default"` only 
        stores `se_root`. `"all"` stores all calculated variables. Use a
        list to specify a custom output set, by default "default".
    export_to_tif : bool, optional
        Return the variables selected with `export_vars` as netCDF (False) 
        or geoTIFF (True), by default False.

    Returns
    -------
    xr.Dataset | dict
        Returns a dataset or a dictionary with variable names as keys and
        lists with paths to geoTIFF files as values, depending on `export_to_tif`.
    """

    log.info("> SE_ROOT").add()

    # Version
    if se_root_version == "v2":
        ETLook = ETLook_v2
        log.info("--> Running SEroot_v2.")
    elif se_root_version == "dev":
        ETLook = ETLook_dev
        log.info("--> Running SEroot_dev.")
    c = ETLook.constants

    # Inputs
    if isinstance(input_data, str):
        ds = xr.open_dataset(input_data)
    else:
        ds = input_data
        input_data = ds.encoding["source"]

    if se_root_version == "dev":
        c.z0m_full = 0.04 + 0.01 * (ds.pixel_size - 30)/(250-30)
        ds["lst_zone_mean"] = lst_zone_mean(ds)

    ds["u_i"] = np.sqrt(ds["v2m_i"]**2 + ds["u2m_i"]**2)

    doy = [int(pd.Timestamp(x).strftime("%j")) for x in ds["time"].values]
    ds["doy"] = xr.DataArray(doy, coords = ds["time"].coords)
    dtime = [pd.Timestamp(x).hour + (pd.Timestamp(x).minute / 60) for x in ds["time"].values]
    ds["dtime"] = xr.DataArray(dtime, coords = ds["time"].coords)
    ds["sc"] = ETLook.solar_radiation.seasonal_correction(ds["doy"])
    ds["decl"] = ETLook.solar_radiation.declination(ds["doy"])
    ds["day_angle"] = ETLook.clear_sky_radiation.day_angle(ds["doy"])

    if "r0_bare" not in list(ds.variables): # TODO add this to pre_se_root and remove from pre_et_look
        ds["r0_bare"] = xr.ones_like(ds["ndvi"]) * c.r0_bare
    if "r0_full" not in list(ds.variables): # TODO add this to pre_se_root and remove from pre_et_look
        ds["r0_full"] = xr.ones_like(ds["ndvi"]) * c.r0_full

    if se_root_version == "dev":
        ds["t_air_i"] = xr.where(ds["t_air_i"] < -270, np.nan, ds["t_air_i"])

    ds["p_air_i"] = ETLook.meteo.air_pressure_kpa2mbar(ds["p_air_i"])
    ds["p_air_0_i"] = ETLook.meteo.air_pressure_kpa2mbar(ds["p_air_0_i"])

    ds["vc"] = ETLook.leaf.vegetation_cover(ds["ndvi"], c.nd_min, c.nd_max, c.vc_pow)

    ds["t_air_k_i"] = ETLook.meteo.air_temperature_kelvin_inst(ds["t_air_i"])
    ds["vp_i"] = ETLook.meteo.vapour_pressure_from_specific_humidity_inst(ds["qv_i"], ds["p_air_i"])
    ds["ad_moist_i"] = ETLook.meteo.moist_air_density_inst(ds["vp_i"], ds["t_air_k_i"])
    ds["ad_dry_i"] = ETLook.meteo.dry_air_density_inst(ds["p_air_i"], ds["vp_i"], ds["t_air_k_i"])
    ds["ad_i"] = ETLook.meteo.air_density_inst(ds["ad_dry_i"], ds["ad_moist_i"])
    ds["u_b_i_bare"] = ETLook.soil_moisture.wind_speed_blending_height_bare(ds["u_i"], c.z0m_bare, c.z_obs, c.z_b)
    ds["lon_rad"] = ETLook.solar_radiation.longitude_rad(ds["lon"])
    ds["lat_rad"] = ETLook.solar_radiation.latitude_rad(ds["lat"])
    ds["ha"] = ETLook.solar_radiation.hour_angle(ds["sc"], ds["dtime"], ds["lon_rad"])
    
    I0 = ETLook.clear_sky_radiation.solar_constant() # TODO this should come from ETLook.constants
    ds["ied"] = ETLook.clear_sky_radiation.inverse_earth_sun_distance(ds["day_angle"])
    ds["h0"] = ETLook.clear_sky_radiation.solar_elevation_angle(ds["lat_rad"], ds["decl"], ds["ha"])
    ds["h0ref"] = ETLook.clear_sky_radiation.solar_elevation_angle_refracted(ds["h0"])
    ds["m"] = ETLook.clear_sky_radiation.relative_optical_airmass(ds["p_air_i"], ds["p_air_0_i"], ds["h0ref"])
    ds["rotm"] = ETLook.clear_sky_radiation.rayleigh_optical_thickness(ds["m"])
    ds["Tl2"] = ETLook.clear_sky_radiation.linke_turbidity(ds["wv_i"], c.aod550_i, ds["p_air_i"], ds["p_air_0_i"])
    ds["G0"] = ETLook.clear_sky_radiation.extraterrestrial_irradiance_normal(I0, ds["ied"])
    ds["B0c"] = ETLook.clear_sky_radiation.beam_irradiance_normal_clear(ds["G0"], ds["Tl2"], ds["m"], ds["rotm"], ds["h0"])
    ds["Bhc"] = ETLook.clear_sky_radiation.beam_irradiance_horizontal_clear(ds["B0c"], ds["h0"])
    ds["Dhc"] = ETLook.clear_sky_radiation.diffuse_irradiance_horizontal_clear(ds["G0"], ds["Tl2"], ds["h0"])
    
    ds["ra_hor_clear_i"] = ETLook.clear_sky_radiation.ra_clear_horizontal(ds["Bhc"], ds["Dhc"])
    ds["emiss_atm_i"] = ETLook.soil_moisture.atmospheric_emissivity_inst(ds["vp_i"], ds["t_air_k_i"])
    
    ds["rn_bare"] = ETLook.soil_moisture.net_radiation_bare(ds["ra_hor_clear_i"], ds["emiss_atm_i"], ds["t_air_k_i"], ds["lst"], ds["r0_bare"])
    ds["rn_full"] = ETLook.soil_moisture.net_radiation_full(ds["ra_hor_clear_i"], ds["emiss_atm_i"], ds["t_air_k_i"], ds["lst"], ds["r0_full"])
    ds["h_bare"] = ETLook.soil_moisture.sensible_heat_flux_bare(ds["rn_bare"], c.fraction_h_bare)
    ds["h_full"] = ETLook.soil_moisture.sensible_heat_flux_full(ds["rn_full"], c.fraction_h_full)
    ds["u_b_i_full"] = ETLook.soil_moisture.wind_speed_blending_height_full_inst(ds["u_i"], c.z0m_full, c.z_obs, c.z_b)
    
    ds["u_star_i_bare"] = ETLook.soil_moisture.friction_velocity_bare_inst(ds["u_b_i_bare"], c.z0m_bare, c.disp_bare, c.z_b)
    ds["u_star_i_full"] = ETLook.soil_moisture.friction_velocity_full_inst(ds["u_b_i_full"], c.z0m_full, c.disp_full, c.z_b)
    ds["L_bare"] = ETLook.soil_moisture.monin_obukhov_length_bare(ds["h_bare"], ds["ad_i"], ds["u_star_i_bare"], ds["t_air_k_i"])
    ds["L_full"] = ETLook.soil_moisture.monin_obukhov_length_full(ds["h_full"], ds["ad_i"], ds["u_star_i_full"], ds["t_air_k_i"])
    
    ds["u_i_soil"] = ETLook.soil_moisture.wind_speed_soil_inst(ds["u_i"], ds["L_bare"], c.z_obs)
    ds["ras"] = ETLook.soil_moisture.aerodynamical_resistance_soil(ds["u_i_soil"])
    ds["raa"] = ETLook.soil_moisture.aerodynamical_resistance_bare(ds["u_i"], ds["L_bare"], c.z0m_bare, c.disp_bare, c.z_obs)
    ds["rac"] = ETLook.soil_moisture.aerodynamical_resistance_full(ds["u_i"], ds["L_full"], c.z0m_full, c.disp_full, c.z_obs)
    
    ds["t_max_bare"] = ETLook.soil_moisture.maximum_temperature_bare(ds["ra_hor_clear_i"], ds["emiss_atm_i"], ds["t_air_k_i"], ds["ad_i"], ds["raa"], ds["ras"], ds["r0_bare"])
    ds["t_max_full"] = ETLook.soil_moisture.maximum_temperature_full(ds["ra_hor_clear_i"], ds["emiss_atm_i"], ds["t_air_k_i"], ds["ad_i"], ds["rac"], ds["r0_full"])

    # od["w_i"] = ETLook.soil_moisture.dew_point_temperature_inst(od["vp_i"])
    # od["t_dew_i"] = ETLook.soil_moisture.dew_point_temperature_inst(od["vp_i"])
    # od["t_wet_i"] = ETLook.soil_moisture.wet_bulb_temperature_inst(id["t_air_i"], od["t_dew_i"], id["p_air_i"])
    ds["t_wet_i"] = ETLook.soil_moisture.wet_bulb_temperature_inst_new(ds["t_air_i"], ds["qv_i"], ds["p_air_i"])
    #t_wet_i_new2 = ETLook.soil_moisture.wet_bulb_temperature_inst_new2(t_air_i)

    ds["lst_max"] = ETLook.soil_moisture.maximum_temperature(ds["t_max_bare"], ds["t_max_full"], ds["vc"])

    if se_root_version == "v2":
        ds["t_wet_k_i"] = ETLook.meteo.wet_bulb_temperature_kelvin_inst(ds["t_wet_i"])
        ds["lst_min"] = ETLook.soil_moisture.minimum_temperature(ds["t_wet_k_i"], ds["t_air_k_i"], ds["vc"])
    elif se_root_version == "dev":
        ds["t_min_bare"] = ETLook.soil_moisture.minimum_temperature_bare(ds["ra_hor_clear_i"], ds["emiss_atm_i"], ds["t_air_k_i"], ds["ad_i"], ds["raa"], ds["ras"], ds["lst_zone_mean"], c.r0_bare_wet)
        ds["t_min_full"] = ETLook.soil_moisture.minimum_temperature_full(ds["ra_hor_clear_i"], ds["emiss_atm_i"], ds["t_air_k_i"], ds["ad_i"], ds["rac"], ds["lst_zone_mean"], ds["r0_full"])  
        ds["lst_min"] = ETLook.soil_moisture.maximum_temperature(ds["t_min_bare"], ds["t_min_full"], ds["vc"])

    ds["se_root"] = ETLook.soil_moisture.soil_moisture_from_maximum_temperature(ds["lst_max"], ds["lst"], ds["lst_min"])

    if export_vars == "all":
        ...
    elif export_vars == "default":
        keep_vars = ['se_root']
        ds = PF.ds_remove_except(ds, keep_vars)
    elif isinstance(export_vars, list):
        keep_vars = copy.copy(export_vars)
        ds = PF.ds_remove_except(ds, keep_vars)
    else:
        raise ValueError

    ds = g.variables.fill_attrs(ds)

    fp, fn = os.path.split(input_data)
    fn = fn.replace("_input", "_output")
    ds, fh = calculate_ds(ds, os.path.join(fp, fn), "--> Saving outputs.")

    log.sub().info("< SE_ROOT")

    if export_to_tif:
        files = PF.export_ds_to_tif(ds, keep_vars, None)
        ds.close()
        # os.remove(fh)
        return files
    else:
        return ds

def lst_zone_mean(ds):
    # TODO This function needs to be replaced by something like pywapor.enhancers.temperature.local_mean

    geo_ex = ds.geotransform
    proj_ex = ds.projection

    size_y, size_x = ds["ndvi"].isel(time = 0).shape
    size_y_zone = int(np.ceil(size_y/200))
    size_x_zone = int(np.ceil(size_x/200)) 
    array_fake = np.ones([size_y_zone, size_x_zone])
    geo_new = tuple([geo_ex[0], geo_ex[1] * 200, geo_ex[2], geo_ex[3], geo_ex[4], geo_ex[5]*200])
    MEM_file = PF.Save_as_MEM(array_fake, geo_new, proj_ex)

    lst_zone_mean_full = np.ones_like(ds["ndvi"] * np.nan)

    for i, t in enumerate(ds.time):

        lst = ds["lst"].sel(time = t).values
        lst_filename = PF.Save_as_MEM(lst, geo_new, proj_ex)
        
        dest_lst_zone_large = PF.reproject_dataset_example(lst_filename, MEM_file, 4)
        lst_zone_mean_large = dest_lst_zone_large.GetRasterBand(1).ReadAsArray()
        lst_zone_mean_large[lst_zone_mean_large==0] = -9999
        lst_zone_mean_large[np.isnan(lst_zone_mean_large)] = -9999

        if np.nanmax(lst_zone_mean_large) == -9999:
            for x in range(0, size_x_zone):
                for y in range(0, size_y_zone):
                    lst_zone_mean_large[y, x] = np.nanmean(lst[y*200:np.minimum((y+1)*200, size_y-1), x*200:np.minimum((x+1)*200, size_x-1)])
            lst_zone_mean_large[np.isnan(lst_zone_mean_large)] = -9999

        lst_zone_mean_large = PF.gap_filling(lst_zone_mean_large, -9999, 1)
        dest_lst_zone_large = PF.Save_as_MEM(lst_zone_mean_large, geo_new, proj_ex)
        
        dest_lst_zone = PF.reproject_dataset_example(dest_lst_zone_large, lst_filename, 6)
        lst_zone_mean = dest_lst_zone.GetRasterBand(1).ReadAsArray()

        lst_zone_mean_full[i, ...] = lst_zone_mean
        
    lst_zone_mean_da = xr.DataArray(lst_zone_mean_full, coords = ds.ndvi.coords)

    return lst_zone_mean_da

if __name__ == "__main__":

    import pywapor

    project_folder = r"/Users/hmcoerver/pywapor_notebooks"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]

    startdate = "2021-07-06"
    enddate = "2021-07-07"
    composite_length = 1
    level = "level_2"
    extra_sources = None
    extra_source_locations = None

    my_custom_level = {
        # Main inputs
        "ndvi":         ["PROBAV", "MOD13", "MYD13"],
        "r0":           ["PROBAV", "MCD43"],
        "lst":          ["MOD11", "MYD11"],
        "lulc":         ["GLOBCOVER"],
        "z":            ["SRTM"],
        "p_24":         ["CHIRPS"],
        "ra_24":        ["MERRA2"],

        # Daily meteo 
        't_air_24':     ["MERRA2", "GEOS5"],
        't_air_min_24': ["MERRA2"], 
        't_air_max_24': ["MERRA2"],
        'u2m_24':       ["GEOS5"],
        'v2m_24':       ["GEOS5"],
        'p_air_0_24':   ["MERRA2"],
        'qv_24':        ["MERRA2", "GEOS5"],

        # Instanteneous meteo
        "t_air_i":      ["MERRA2"],
        "u2m_i":        ["MERRA2"],
        "v2m_i":        ["MERRA2"],
        "qv_i":         ["MERRA2"],
        "wv_i":         ["MERRA2"],
        "p_air_i":      ["MERRA2"],
        "p_air_0_i":    ["MERRA2"],

        # Temporal constants
        "lw_offset":    ["STATICS"],
        "lw_slope":     ["STATICS"],
        "r0_bare":      ["STATICS"],
        "r0_full":      ["STATICS"],
        "rn_offset":    ["STATICS"],
        "rn_slope":     ["STATICS"],
        "t_amp_year":   ["STATICS"],
        "t_opt":        ["STATICS"],
        "vpd_slope":    ["STATICS"],
        "z_oro":        ["STATICS"],

        # Level name
        "name": "test",
    }

    diagnostics = None

    ds, fh = pywapor.pre_se_root.main(project_folder, startdate, enddate, latlim, 
                                        lonlim, level = my_custom_level,
                                        extra_sources = extra_sources, 
                                        extra_source_locations = extra_source_locations)

    se_root_version = "v2"

    ds_out = main(ds, se_root_version = se_root_version, export_vars = "all", export_to_tif = False)


    


# -*- coding: utf-8 -*-
"""
Generates input data for `pywapor.se_root`.
"""
import warnings
import os
import copy
import xarray as xr
import numpy as np
import pandas as pd
import pywapor.general as g
from pywapor.collect.downloader import collect_sources
from pywapor.general.logger import log, adjust_logger
from datetime import datetime as dat
from datetime import time as datt
from pywapor.general.compositer import check_geots, preprocess_func
from pywapor.enhancers.temperature import kelvin_to_celsius
from pywapor.enhancers.apply_enhancers import apply_enhancer
from pywapor.general.compositer import calculate_ds
from pywapor.general import processing_functions as pf

def main(project_folder, startdate, enddate, latlim, lonlim, level = "level_1", 
            extra_source_locations = None):
    """Generates an input file for `pywapor.se_root` based on, among others, a
    bounding-box and time period.

    Parameters
    ----------
    project_folder : str
        Path to folder in which (intermediate) data will be stored.
    startdate : str
        Start of the period for which to generate the input data, formatted as
        "YYYY-MM-DD".
    enddate : str
        End of the period for which to generate the input data, formatted as
        "YYYY-MM-DD".
    latlim : List[float, float]
        Latitude limits of the bounding-box, first value should be smaller than
        second value.
    lonlim : List[float, float]
        Longitude limits of the bounding-box, first value should be smaller than
        second value.
    level : "level_1" | "level_2" | dict, optional
        Controls which sources are collected for the various variables, 
        by default "level_1".
    extra_source_locations : dict, optional
        Paths in which to look for sideloaded data, by default {}.

    Returns
    -------
    xr.Dataset
        Dataset with all the required data for pywapor.se_root.
    str
        Path to the netCDF file containing the generated dataset.

    """

#%%

    log_write = True
    log_level = "INFO"
    adjust_logger(log_write, project_folder, log_level)

    log.info("> PRE_SE_ROOT").add()

    # Disable Warnings
    warnings.filterwarnings('ignore')

    # Load required variable sources.
    levels = g.variables.get_source_level_selections()

    if isinstance(level, str):
        source_selection = levels[level]
        level_name = level
    elif isinstance(level, dict):
        source_selection = copy.copy(level)
        if "level_name" in source_selection.keys():
            level_name = source_selection.pop("level_name")
        else:
            level_name = "custom"

    raw_folder = os.path.join(project_folder, "RAW")
    level_folder = os.path.join(project_folder, level_name)
    temp_folder = os.path.join(project_folder, "temporary")

    dl_args = {
                "Dir": raw_folder, 
                "latlim": latlim, 
                "lonlim": lonlim, 
                "Startdate": startdate, 
                "Enddate": enddate,
                }

    #### NDVI ####
    log.info(f"# ndvi")
    raw_ndvi_files = collect_sources("ndvi", source_selection["ndvi"], dl_args, extra_source_locations)

    example_fh, example_ds, example_geoinfo, resolution = pf.select_template(raw_ndvi_files)

    cmeta = {
        "composite_type": None, # 
        "temporal_interp": False,
        "temporal_interp_freq": 1,
        "spatial_interp": "nearest",
        "var_name": "ndvi",
        "var_unit": "-",
    }
    ds_ndvi = g.compositer.main(cmeta, raw_ndvi_files, None, temp_folder, None, lean_output = False)
    ds_ndvi = ds_ndvi.drop_vars(["sources"]).chunk("auto")

#%%
    #### LST ####
    log.info("# lst")
    raw_lst_files = collect_sources("lst", source_selection["lst"], dl_args, extra_source_locations)

    # TODO `combine_lst` needs to go into collect, then this whole block can go away!
    raw_modis_files = [x for x in raw_lst_files if isinstance(x, dict)]
    raw_nonmodis_files = [x for x in raw_lst_files if isinstance(x, list)]

    if len(raw_modis_files) > 0:
        ds_lst = combine_lst(raw_lst_files) 

        ds_lst2a = ds_lst.interp_like(example_ds, method = "linear", kwargs={"fill_value": "extrapolate"},)
        fh = os.path.join(temp_folder, "lst_ts.nc")
        ds_lst2a, _ = calculate_ds(ds_lst2a, fh, label = "--> Resampling datasets.")
    else:
        ds_lst2a = False

    if len(raw_nonmodis_files) > 0:
        cmeta = {
            "composite_type": None, # 
            "temporal_interp": False,
            "temporal_interp_freq": 1,
            "spatial_interp": "nearest",
            "var_name": "lst",
            "var_unit": "-",
        }
        ds_lst2b = g.compositer.main(cmeta, raw_nonmodis_files, None, temp_folder, example_ds, lean_output = False)
        ds_lst2b = ds_lst2b.drop_vars(["sources"]).chunk("auto")
    else:
        ds_lst2b = False

    if ds_lst2a and ds_lst2b:
        ds_lst2 = xr.merge([ds_lst2a, ds_lst2b])
    elif ds_lst2a:
        ds_lst2 = ds_lst2a
    elif ds_lst2b:
        ds_lst2 = ds_lst2b
    else:
        raise ValueError

#%%
    #### METEO ####
    log.info("> METEO").add()

    all_vars = ["t_air_i", "u2m_i", "v2m_i", "qv_i", 
                "wv_i", "p_air_i", "p_air_0_i"]

    dss = list()

    meteo_enhancements = {
        ("MERRA2",  "t_air_i"):        [kelvin_to_celsius],
        ("GEOS5",   "t_air_i"):        [kelvin_to_celsius],
    }

    for var in all_vars:

        log.info(f"# {var}")

        meteo_source = source_selection[var][0]

        if len(source_selection[var]) > 1:
            log.warning(f"! --> Multiple sources not supported for '{var}', using '{meteo_source}' only.")

        meteo_freq = {"GEOS5": "3H", "MERRA2": "H"}[meteo_source]

        sds, eds, prds = calc_periods(ds_lst2.time.values, meteo_freq)

        dl_args["Startdate"] = sds
        dl_args["Enddate"] = eds
        dl_args["Periods"] = prds

        db = collect_sources(var, [meteo_source], dl_args, 
                                extra_source_locations = extra_source_locations)[0]

        check_geots(db)
        ds = xr.open_mfdataset(db, concat_dim = "time", engine="rasterio", 
                                combine = "nested", preprocess = preprocess_func)
        ds = ds.rename({"band_data": var})

        if (meteo_source, var) in meteo_enhancements.keys():
            for enhancer in meteo_enhancements[(meteo_source, var)]:
                ds, label = apply_enhancer(ds, var, enhancer, 
                                                source = meteo_source)
                ds, _ = calculate_ds(ds, label = label)
        else:
            ds, _ = calculate_ds(ds, fh = os.path.join(temp_folder, f"{var}_ts.nc"))

        ds = ds.chunk("auto")

        attributes = {
                "sources": [meteo_source],
            }
        ds[var].attrs = attributes

        # temporal
        ds, x, y = g.compositer.temporal_interpolation(ds, ds_lst2.time.values, "nearest")
        ds = ds.sel(time = ds_lst2.time.values)

        # spatial
        ds = ds.interp_like(example_ds, method = "linear", kwargs={"fill_value": "extrapolate"},)
        
        dss.append(ds)

    ds_meteo3 = xr.merge(dss, combine_attrs = "drop_conflicts")

    log.sub().info("< METEO")

    ds_ndvi3, x, y = g.compositer.temporal_interpolation(ds_ndvi, ds_lst2.time.values, "linear")
    ds_ndvi3 = ds_ndvi3.sel(time = ds_lst2.time)

    ds_se_root = xr.merge([ds_lst2, ds_meteo3, ds_ndvi3])

    if "spatial_ref" in list(ds_se_root.coords):
        ds_se_root = ds_se_root.drop_vars("spatial_ref")
    if "band" in list(ds_se_root.coords):
        ds_se_root = ds_se_root.drop_vars("band")

    ds_se_root = g.variables.fill_attrs(ds_se_root)

    ds_se_root.attrs["geotransform"] = example_geoinfo[0]
    ds_se_root.attrs["projection"] = example_geoinfo[1]
    ds_se_root.attrs["pixel_size"] = resolution
    ds_se_root.attrs["example_file"] = example_fh

    fh = os.path.join(level_folder, "se_root_input.nc")
    ds, fh = calculate_ds(ds_se_root, fh, "--> Interpolating datasets.")

    log.sub().info("< PRE_SE_ROOT")

    return ds, fh

def calc_periods(times, freq):
    """Calculates which GEOS5 or MERRA2 periods for a given time need to be collected,
    e.g. checks which GEOS5 or MERRA2 datapoint in time is closest to the given
    `times`. Used by `pywapor.Collect.MERRA2` and `pywapor.Collect.GEOS5`.

    Parameters
    ----------
    times : np.ndarray
        Array with values that can be ingested by pd.Timestamp.
    freq : {"3H" | "H"}
        The frequency of the data for which the periods are calculated.

    Returns
    -------
    list
        List of start datetimes.
    list
        List of end datetimes.
    list
        List of periods corresponding to the gives `times`.
    """
    # TODO move this to collect tools?
    sds = list()
    eds = list()
    prds = list()

    for now_time in times:

        date = pd.Timestamp(now_time).date()
        starttime = dat(date.year, date.month, date.day, 0, 0)
        endtime = dat(date.year, date.month, date.day, 23, 59)

        offsets = {"3H": 90, "H": 30}

        DateTime = pd.date_range(starttime, endtime, freq=freq) + pd.offsets.Minute(offsets[freq])
        Time_nearest = min(DateTime, key=lambda DateTime: abs(DateTime - now_time))
        period = np.argwhere(DateTime ==Time_nearest)[0][0] + 1

        sds.append(starttime)
        eds.append(endtime)
        prds.append([period])

    return sds, eds, prds

def combine_lst(lst_files):
    ds = None

    sources = list()

    for lst_db in lst_files:

        fn = os.path.split(list(lst_db.values())[0][0])[-1]
        source = fn.split("_")[1]
        sources.append(source)

        for date, files in lst_db.items():

            ds_time = xr.open_dataset(files[1]).squeeze("band").rename({"band_data": "time", "x": "lon", "y": "lat"})
            ds_angle = xr.open_dataset(files[2]).squeeze("band").rename({"band_data": "angle", "x": "lon", "y": "lat"})
            ds_lst = xr.open_dataset(files[0]).squeeze("band").rename({"band_data": "lst", "x": "lon", "y": "lat"})

            unique_times = np.unique(ds_time.time)
            unique_times = unique_times[np.isfinite(unique_times)]

            offset_GTM = int(round(np.median(ds_time.lon) * 24 / 360)) # TODO Make this spatial for areas spanning multiple timezones

            for time in unique_times:

                hours = np.floor(time).astype(int) - offset_GTM
                minutes = np.round((time - offset_GTM - hours) * 60).astype(int)
                dt = dat.combine(date, datt(hours, minutes))

                da_angle = ds_angle.angle.where(ds_time.time == time)
                da_angle = da_angle.assign_coords({"time": dt}).expand_dims("time", axis = 0)

                da_lst = ds_lst.lst.where(ds_time.time == time, np.nan)
                da_lst = da_lst.assign_coords({"time": dt}).expand_dims("time", axis = 0)

                if isinstance(ds, type(None)):
                    ds = xr.Dataset({"angle": da_angle, "lst": da_lst})
                else:
                    ds_temp = xr.Dataset({"angle": da_angle, "lst": da_lst})
                    ds = xr.concat([ds, ds_temp], dim = "time")

    ds["lst"].attrs = {"sources": sources}
    ds["angle"].attrs = {"sources": sources}

    return ds

if __name__ == "__main__":

    project_folder = r"/Users/hmcoerver/pywapor_notebooks"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    startdate = "2021-07-01"
    enddate = "2021-07-11"
    composite_length = "DEKAD"
    # level = "level_2"
    # extra_source_locations = None

    level = {
        # Main inputs
        "ndvi":         ["LS7"],
        "r0":           ["LS7"],
        "lst":          ["LS7"],
        "lulc":         ["WAPOR"],
        "z":            ["SRTM"],
        "p_24":         ["CHIRPS"],
        "ra_24":        ["MERRA2"],

        # Daily meteo 
        't_air_24':     ["GEOS5"],
        't_air_min_24': ["GEOS5"], 
        't_air_max_24': ["GEOS5"],
        'u2m_24':       ["GEOS5"],
        'v2m_24':       ["GEOS5"],
        'p_air_0_24':   ["GEOS5"],
        'qv_24':        ["GEOS5"],

        # Instanteneous meteo
        "t_air_i":      ["GEOS5"],
        "u2m_i":        ["GEOS5"],
        "v2m_i":        ["GEOS5"],
        "qv_i":         ["GEOS5"],
        "wv_i":         ["GEOS5"],
        "p_air_i":      ["GEOS5"],
        "p_air_0_i":    ["GEOS5"],

        # Temporal constants
        "lw_offset":    ["STATICS"],
        "lw_slope":     ["STATICS"],
        "r0_bare":      ["STATICS"],
        "r0_full":      ["STATICS"],
        "rn_offset":    ["STATICS"],
        "rn_slope":     ["STATICS"],
        "t_amp_year":   ["STATICS"],
        "t_opt":        ["STATICS"],
        "vpd_slope":    ["STATICS"],
        "z_oro":        ["STATICS"],

        # Level name
        "level_name": "sideloading",
    }

    # Give product folder.
    extra_source_locations = {
        ("LS7", "ndvi"): r"/Users/hmcoerver/pywapor_notebooks/my_landsat_folder/NDVI",
        ("LS7", "lst"): r"/Users/hmcoerver/pywapor_notebooks/my_landsat_folder/LST",
        ("LS7", "r0"): r"/Users/hmcoerver/pywapor_notebooks/my_landsat_folder/ALBEDO",
    }

    main(project_folder, startdate, enddate, latlim, lonlim, level = level,
        extra_source_locations = extra_source_locations)


# -*- coding: utf-8 -*-
"""

Description:
This module downloads MYD13 NDVI data from
http://e4ftl01.cr.usgs.gov/. Use the MYD13.NDVI function to
download and create 16 daily NDVI images in Gtiff format.
The data is available between 2000-02-18 till present.

Examples:
from pyWAPOR.Collect import MYD13
MYD13.NDVI(Dir='C:/Temp3/', Startdate='2003-12-01', Enddate='2003-12-20',
           latlim=[41, 45], lonlim=[-8, -5])
"""

from pywapor.collect.MYD13.NDVI import main as NDVI

__all__ = ['NDVI']

__version__ = '0.1'



import glob
import os
from pywapor.general.logger import log

def search_product_files(product_name, path):
    """Search for geotiff files with a specific `product_name` in a folder
    defined by `path`.

    Parameters
    ----------
    product_name : str
        Should be inside a geotiffs file's name after one underscore.
    path : str
        Path to folder to search.

    Return
    ------
    list
        Paths to files with the product_name inside their filename.

    """
    fps = glob.glob(os.path.join(path, f"*_{product_name}_*.tif"))
    log.info(f"--> Collected {len(fps)} {product_name} file(s).")
    check = {os.path.split(fp)[-1]: check_filename(os.path.split(fp)[-1]) for fp in fps}
    log.debug(f"Filenames are incorrect ({check}).")
    return fps

def check_filename(fn):
    """Checks if files follow the correct naming conventions.         
    
    Filenames should look like this:
    `{variable_name}_{product_name}_{unit}_{period_length}_{date}.tif`
    where `variable_name`, `product_name` and `unit` are strings without 
    underscores. `period_length` can be either `-` (for instanteneous 
    data), an integer value or an integer value appended with "-daily". 
    `date` is either formatted as `%Y.%m.%d` in case a `period_length` is 
    specified OR `%Y.%m.%d.%H.%M` for instantenous data.

    Parameters
    ----------
    fn : str
        Filename to test.

    Returns
    -------
    bool
        True when the filename is correct.

    Examples
    --------
    >>> check_filenames("NDVI_MOD13_-_16-daily_2021.06.10")
    True
    >>> check_filenames("NDVI_MOD13_-_16_2021.06.10")
    True
    >>> check_filenames("ND_VI_MOD13_-_16_2021.06.10")
    False
    >>> check_filenames("NDVI_MOD13_-_16_2021.06.18.00.00")
    False
    >>> check_filenames("NDVI_MOD13_-_-_2021.06.18.00.00")
    True
    """
    name, ext = os.path.splitext(fn)
    parts = name.split("_")

    if ext != ".tif":
        raise ValueError("File has a wrong extension ({ext} != '.tif').")

    if len(parts) != 5:
        raise ValueError("Too many underscores in filename.")

    date_parts = parts[4].split(".")

    if len(date_parts) == 3:
        if "daily" in parts[3]:
            period_length = int(parts[3].split("-")[0])
            if period_length - float(parts[3].split("-")[0]) != 0:
                raise ValueError(f"Incorrect period_length ({period_length}).")
        else:
            if int(parts[3]) - float(parts[3]) != 0:
                raise ValueError(f"Incorrect period_length ({parts[3]}).")
    elif len(date_parts) == 5:
        if parts[3] != "-":
            raise ValueError(f"Incorrect period_length ({parts[3]}). Should be '-'.")
    else:
        raise ValueError(f"Invalid date ({parts[4]}).")

    return True



# -*- coding: utf-8 -*-
from pywapor.collect.GEOS5.DataAccess import DownloadData
import tqdm
import pandas as pd
from pywapor.general.logger import log

def main(Dir, latlim, lonlim, Startdate, Enddate, Vars, Waitbar = True):
    """
    This function downloads GEOS daily data for a given variable, time
    interval, and spatial extent.

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Vars -- ['t2m', 'v2m'] 
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
    Waitbar -- 1 (Default) Will print a waitbar
    """
    no_dates = len(pd.date_range(Startdate, Enddate, freq="D"))
    no_vars = len(Vars)

    if len(Vars) == 1:
        log.info(f"--> Downloading GEOS5 (daily), {Vars[0]}.")
    else:
        log.info(f"--> Downloading GEOS5 (daily), {Vars}.")

    if Waitbar:
        # waitbar = tqdm.tqdm(total = no_vars * no_dates, position = 0, unit = "tiles")

        waitbar = tqdm.tqdm(desc= f"Tile: 0 / {no_vars * no_dates}",
                            position = 0,
                            # total=total_size,
                            unit='Bytes',
                            unit_scale=True,)

    else:
        waitbar = False

    all_files = list()
    for Var in Vars:

        if Var == "t2m":
            filter = "t2m_"
        elif Var == "t2m-max":
            Var = "t2m"
            filter = "t2m-max_"
        elif Var == "t2m-min":
            Var = "t2m"
            filter = "t2m-min_"
        else:
            filter = None

        # if Waitbar:
        #     waitbar.set_description_str("{:<11}".format(Var))

        # Download data
        new_files = DownloadData(Dir, Var, Startdate, Enddate, latlim, lonlim, "daily", Period = '', Waitbar = waitbar)
        
        if not isinstance(filter, type(None)):
            new_files = [x for x in new_files if filter in x]
        
        all_files = all_files + new_files

    if Waitbar:
        waitbar.close()

    return all_files

if __name__ == '__main__':
    
    Dir = r"/Users/hmcoerver/pywapor_notebooks/RAW"
    latlim = [-40.0, 20.0]
    lonlim = [80.0, 100.0]
    Startdate = "2021-07-01"
    Enddate = "2021-07-03"

    Vars = ['u2m']

    main(Dir, latlim, lonlim, Startdate, Enddate, Vars, Waitbar = True)



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: Collect/SRTM
"""
from pywapor.collect.WAPOR.DataAccess import DownloadData
import sys

def main(Dir, Startdate, Enddate, latlim, lonlim, Waitbar = 1):
    """
    Downloads LandCover data from https://io.apps.fao.org/gismgr/api/v1/query/.

    The following keyword arguments are needed:
    Dir -- 'C:/file/to/path/'
	Startdate -- 'yyyy-mm-dd'
	Enddate -- 'yyyy-mm-dd'    
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
    Waitbar -- '1' if you want a waitbar (Default = 1)
    """

    print('\nDownloading Yearly LandCover maps from WAPOR for the period %s till %s' %(Startdate, Enddate))

    # Download data
    DownloadData(Dir, Startdate, Enddate, latlim, lonlim, Waitbar)

if __name__ == '__main__':
    main(sys.argv)




import os
import numpy as np
import pandas as pd
import datetime
import requests
from netCDF4 import Dataset
import tqdm
from pywapor.general.logger import log
import time

def DownloadData(Dir, Var, Startdate, Enddate, latlim, lonlim, TimeStep, Period, username, password, Waitbar, data_type = ["mean"]):

    import pywapor.general.processing_functions as PF

    all_files = list()
    # if "mean" in data_type:
    #     all_files[Var] = list()
    # if "max" in data_type:
    #     all_files[f"{Var}-max"] = list()
    # if "min" in data_type:
    #     all_files[f"{Var}-min"] = list()

    # Add extra buffer to ensure good spatial interpolation
    buffer_pixels = 0
    lonlim = [lonlim[0] - 0.625 * buffer_pixels, lonlim[1] + 0.625 * buffer_pixels]
    latlim = [latlim[0] - 0.5 * buffer_pixels, latlim[1] + 0.5 * buffer_pixels]

    # Check the latitude and longitude and otherwise set lat or lon on greatest extent
    if latlim[0] < -90 or latlim[1] > 90:
        log.warning('Latitude above 90N or below 90S is not possible. Value set to maximum')
        latlim[0] = np.max(latlim[0], -90)
        latlim[1] = np.min(latlim[1], 90)
    if lonlim[0] < -180 or lonlim[1] > 180:
        log.warning('Longitude must be between 180E and 180W. Now value is set to maximum')
        lonlim[0] = np.max(lonlim[0], -180)
        lonlim[1] = np.min(lonlim[1], 180)  
    
    # Get information of the parameter    
    VarInfo = VariablesInfo(TimeStep)
    Parameter = VarInfo.names[Var]
    unit  = VarInfo.units[Var]
    p_ts = VariablesInfo.period_times

    # Create output folder
    output_folder = os.path.join(Dir, "MERRA2", Parameter, TimeStep) 
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    
    corrx = 0.625 * 0.5
    corry = 0.5 * 0.5

    # Define IDs
    IDx = [np.floor((lonlim[0] + corrx + 180)/0.625), np.ceil((lonlim[1] + corrx + 180)/0.625)]
    IDy = [np.floor((latlim[0] + corry + 90)/0.5), np.ceil((latlim[1] + corry + 90)/0.5)]
    
    # Create output geo transform
    Xstart = -180 + 0.625 * IDx[0] - corrx
    Ystart = -90 + 0.5 * IDy[1] - corry
    
    geo_out = tuple([Xstart, 0.625, 0, Ystart, 0, -0.5])
    proj = "WGS84"
    
    Dates = pd.date_range(Startdate, Enddate, freq = "D")

    if Waitbar:
        waitbar = tqdm.tqdm(desc= f"Tile: 0 / {len(Dates)}",
                            position = 0,
                            # total=total_size,
                            unit='Bytes',
                            unit_scale=True,)
    else:
        waitbar = None

    for Date in Dates:
        
        # Define the IDz
        if TimeStep == "hourly_MERRA2":
            if Period < 1 or Period > 24:
                raise ValueError(f"Invalid Period (={Period}).")

            date = datetime.datetime.combine(Date, 
                                    p_ts[Period])
            
            fn = f"{Var}_MERRA2_{unit}_inst_{date:%Y.%m.%d.%H.%M}.tif"
            output_name = os.path.join(output_folder, fn)

            output_folder_temp = os.path.join(Dir, "MERRA2", "Temp")
            if not os.path.exists(output_folder_temp):
                os.makedirs(output_folder_temp)
            year = Date.year
            month = Date.month
            day = Date.day
            output_name_min = output_folder
            output_name_max = output_folder
            if "min" in data_type:
                log.warning(f"data_type = {data_type} not applicable to hourly_MERRA2, ignoring.")
                data_type.remove("min")
            if "max" in data_type:
                log.warning(f"data_type = {data_type} not applicable to hourly_MERRA2, ignoring.")
                data_type.remove("max")
            
        if TimeStep == "daily_MERRA2":

            if "mean" in data_type:
                output_name = os.path.join(output_folder, "%s_MERRA2_%s_daily_%d.%02d.%02d.tif" %(Var, unit, Date.year, Date.month, Date.day))
            else:
                output_name = output_folder
            if "min" in data_type:
                output_name_min = os.path.join(output_folder, "%smin_MERRA2_%s_daily_%d.%02d.%02d.tif"%(Var, unit, Date.year, Date.month, Date.day))
            else:
                output_name_min = output_folder
            if "max" in data_type:
                output_name_max = os.path.join(output_folder, "%smax_MERRA2_%s_daily_%d.%02d.%02d.tif"%(Var, unit, Date.year, Date.month, Date.day))
            else:
                output_name_max = output_folder
                
            output_folder_temp = os.path.join(Dir, "MERRA2", "Temp")
            if not os.path.exists(output_folder_temp):
                os.makedirs(output_folder_temp)            
            year = Date.year
            month = Date.month
            day = Date.day
    
        if not (os.path.exists(output_name) and os.path.exists(output_name_min) and os.path.exists(output_name_max)):

            # TODO this shouldnt be hardcoded, cause it will keep giving problems going forward.
            if Date < datetime.datetime(1992,1,1):
                number = 1
            elif (Date >= datetime.datetime(1992,1,1) and Date < datetime.datetime(2001,1,1)):
                number = 2
            elif (Date >= datetime.datetime(2001,1,1) and Date < datetime.datetime(2011,1,1)):
                number = 3
            else:
                number = 4
            if Date.month == 9 and Date.year == 2020:
                number2 = 1
            elif Date.month == 6 and Date.year == 2021:
                number2 = 1
            elif Date.month == 7 and Date.year == 2021:
                number2 = 1
            elif Date.month == 8 and Date.year == 2021:
                number2 = 1
            elif Date.month == 9 and Date.year == 2021:
                number2 = 1
            else:
                number2 = 0
                            
            if Var == "swgnet":
                url_MERRA = r"https://goldsmr4.gesdisc.eosdis.nasa.gov/data/MERRA2/M2T1NXRAD.5.12.4/%d/%02d/MERRA2_%s0%s.tavg1_2d_rad_Nx.%d%02d%02d.nc4" %(year, month, number, number2, year, month, day)
            else:    
                url_MERRA = r"https://goldsmr4.gesdisc.eosdis.nasa.gov/data/MERRA2/M2I1NXASM.5.12.4/%d/%02d/MERRA2_%s0%s.inst1_2d_asm_Nx.%d%02d%02d.nc4" %(year, month, number, number2, year, month, day)
                
            # Reset the begin parameters for downloading
            downloaded = 0
            N = 0

            # if not downloaded try to download file
            while downloaded == 0:

                # Define the output name that is downloaded
                file_name = os.path.join(output_folder_temp, url_MERRA.split("/")[-1])

                if os.path.exists(file_name):
                    done = os.stat(file_name).st_size > 1000
                    if not done:
                        os.remove(file_name)
                else:
                    done = False

                if not done:
                
                    # make contact with server
                    x = requests.get(url_MERRA, allow_redirects = False)

                    x.raise_for_status()

                    resp = requests.get(x.headers['location'], auth = (username, password), stream=True)
                    total_size = int(resp.headers.get('content-length', 0))
                    
                    try:
                        resp.raise_for_status()
                    except requests.exceptions.HTTPError as e:
                        if N < 5:

                            N += 1
                            log.warning(f"N = {N}, {url_MERRA}")
                            time.sleep(5)
                            continue
                        else:
                            raise e

                    if not isinstance(waitbar, type(None)):
                        waitbar.reset(total = total_size)
                        waitbar_i = int(waitbar.desc.split(" ")[1])
                        waitbar_desc = str(waitbar.desc)
                        waitbar.set_description_str(waitbar_desc.replace(f": {waitbar_i} /", f": {waitbar_i+1} /"))
                    with open(file_name, 'wb') as z:
                        for data in resp.iter_content(chunk_size=1024):
                            size = z.write(data)
                            if not isinstance(waitbar, type(None)):
                                waitbar.update(size)
                                
                    statinfo = os.stat(file_name)
                    # Say that download was succesfull
                    if int(statinfo.st_size) > 1000:
                            downloaded = 1  

                else:
                    downloaded = 1
                    if not isinstance(waitbar, type(None)):
                        waitbar_i = int(waitbar.desc.split(" ")[1])
                        waitbar_desc = str(waitbar.desc)
                        waitbar.set_description_str(waitbar_desc.replace(f": {waitbar_i} /", f": {waitbar_i+1} /")) 

            data_end, data_min, data_max = Get_NC_data_end(file_name,Var, TimeStep, Period, IDy, IDx, VarInfo)

            # Add the VarFactor
            if VarInfo.factors[Var] < 0:
                if "mean" in data_type:                       
                    data_end[data_end != -9999] = data_end[data_end != -9999] + VarInfo.factors[Var]
                    data_end[data_end < -9999] = -9999
                    data_end = np.flipud(data_end)
                if "min" in data_type:
                    data_min[data_min != -9999] = data_min[data_min != -9999] + VarInfo.factors[Var]
                    data_min[data_min < -9999] = -9999
                    data_min = np.flipud(data_min)                                 
                if "max" in data_type:
                    data_max[data_max != -9999] = data_max[data_max != -9999] + VarInfo.factors[Var]      
                    data_max[data_max < -9999] = -9999
                    data_max = np.flipud(data_max)                        
                                            
            else:
                if "mean" in data_type: 
                    data_end[data_end != -9999] = data_end[data_end != -9999] * VarInfo.factors[Var]
                    data_end[data_end < -9999] = -9999
                    data_end = np.flipud(data_end)
                if "min" in data_type:
                    data_min[data_min != -9999] = data_min[data_min != -9999] * VarInfo.factors[Var]
                    data_min[data_min < -9999] = -9999
                    data_min = np.flipud(data_min)                            
                if "max" in data_type:
                    data_max[data_max != -9999] = data_max[data_max != -9999] * VarInfo.factors[Var]                          
                    data_max[data_max < -9999] = -9999
                    data_max = np.flipud(data_max)                        
                
            # Save as tiff file
            if "mean" in data_type: 
                PF.Save_as_tiff(output_name, data_end, geo_out, proj)
                all_files.append(output_name)
            if "min" in data_type:
                PF.Save_as_tiff(output_name_min, data_min, geo_out, proj)
                all_files.append(output_name_min)
            if "max" in data_type:
                PF.Save_as_tiff(output_name_max, data_max, geo_out, proj)
                all_files.append(output_name_max)

        else:
            if os.path.isfile(output_name):
                all_files.append(output_name)
            if os.path.isfile(output_name_min):
                all_files.append(output_name_min)
            if os.path.isfile(output_name_max):
                all_files.append(output_name_max)

            if not isinstance(waitbar, type(None)):
                waitbar_i = int(waitbar.desc.split(" ")[1])
                waitbar_desc = str(waitbar.desc)
                waitbar.set_description_str(waitbar_desc.replace(f": {waitbar_i} /", f": {waitbar_i+1} /")) 

    return all_files  

def Get_NC_data_end(file_name,Var, TimeStep, Period, IDy, IDx, VarInfo):
                        
    dict_para = {'t2m': 'T2M',
         'u2m': 'U2M',
         'v2m': 'V2M',
         'q2m': 'QV2M',
         'tpw': 'TQV',
         'ps': 'PS',
         'slp': 'SLP',
         'swgnet': 'SWGDN'}  
      
    types  = VarInfo.types[Var]
    if TimeStep == "hourly_MERRA2":
        data_end = Dataset(file_name)["%s" %dict_para[Var]][int(Period-1),int(IDy[0]):int(IDy[1]),int(IDx[0]):int(IDx[1])]
        data_min = []
        data_max = []
        
    else:
        data = Dataset(file_name)["%s" %dict_para[Var]][:,int(IDy[0]):int(IDy[1]),int(IDx[0]):int(IDx[1])]
        if types == "state":
            data_end = np.nanmean(data, 0)
        else:
            data_end = np.nansum(data, 0)
            
        data[data==-9999] = np.nan
        data_min = np.nanmin(data, 0)
        data_max = np.nanmax(data, 0)     
        
    return(data_end, data_min, data_max)

class VariablesInfo:
    """
    This class contains the information about the GLDAS variables
    """
    names = {'t2m': 'Air_Temperature',
             'u2m': 'Eastward_Wind',
             'v2m': 'Northward_Wind',
             'q2m': 'Specific_Humidity',
             'tpw': 'Total_Precipitable_Water_Vapor',
             'ps': 'Surface_Pressure',
             'slp': 'Sea_Level_Pressure',
             'swgnet': 'Surface_Net_Downward_Shortwave_Flux'
             }
    
    descriptions = {'t2m': '2m Air Temperature',
             'u2m': '2m Eastward wind',
             'v2m': '2m Northward wind',
             'q2m': '2m Specific Humidity',
             'tpw': 'Total Precipitable Water Vapor',
             'ps': 'Surface Pressure',
             'slp': 'Sea Level Pressure',
             'swgnet': 'Surface Net Downward Shortwave Flux'
             }
    
    factors = {'t2m': 1,
             'u2m': 1,
             'v2m': 1,
             'q2m': 1,
             'tpw': 1,
             'ps': 0.001,
             'slp':  0.001,
             'swgnet':  1
             }
    
    types = {'t2m': 'state',
             'u2m': 'state',
             'v2m': 'state',
             'q2m': 'state',
             'tpw': 'state',
             'ps': 'state',
             'slp': 'state',
             'swgnet': 'state'
             }

    period_times = {1: datetime.time(0, 30),
                    2: datetime.time(1, 30),
                    3: datetime.time(2, 30),
                    4: datetime.time(3, 30),
                    5: datetime.time(4, 30),
                    6: datetime.time(5, 30),
                    7: datetime.time(6, 30),
                    8: datetime.time(7, 30),
                    9: datetime.time(8, 30),
                    10: datetime.time(9, 30),
                    11: datetime.time(10, 30),
                    12: datetime.time(11, 30),
                    13: datetime.time(12, 30),
                    14: datetime.time(13, 30),
                    15: datetime.time(14, 30),
                    16: datetime.time(15, 30),
                    17: datetime.time(16, 30),
                    18: datetime.time(17, 30),
                    19: datetime.time(18, 30),
                    20: datetime.time(19, 30),
                    21: datetime.time(20, 30),
                    22: datetime.time(21, 30),
                    23: datetime.time(22, 30),
                    24: datetime.time(23, 30)}

    def __init__(self, step):
            
        if step == 'hourly_MERRA2':
            self.units = {'t2m': 'K',
             'u2m': 'm-s-1',
             'v2m': 'm-s-1',
             'q2m': 'kg-kg-1',
             'tpw': 'mm',
             'ps': 'kpa',
             'slp': 'kpa',
             'swgnet': 'W-m-2'
             }    
            
        elif step == 'daily_MERRA2':
            self.units = {'t2m': 'K',
             'u2m': 'm-s-1',
             'v2m': 'm-s-1',
             'q2m': 'kg-kg-1',
             'tpw': 'mm',
             'ps': 'kpa',
             'slp': 'kpa',
             'swgnet': 'W-m-2'
             }

        else:
            raise KeyError("The input time step is not supported")

if __name__ == "__main__":

    import pywapor

    Dir = r"/Users/hmcoerver/pywapor_notebooks"
    Startdate = "2021-06-01"
    Enddate = "2021-06-02"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    username, password = pywapor.collect.get_pw_un.get("NASA")
    
    Period = 1
    
    data_type = ["mean", "min", "max"]

    print("daily")
    TimeStep = "daily_MERRA2"
    meteo_vars = ['t2m', 'u2m', 'v2m', 'q2m', 'tpw', 'ps', 'slp', 'swgnet']
    # meteo_vars = ["swgnet"]
    for Var in meteo_vars:
        print(Var)
        all_files1 = DownloadData(Dir, Var, Startdate, Enddate, latlim, lonlim, TimeStep, "", username, password, Waitbar = True, data_type = data_type)

    # print("hourly")
    # TimeStep = "hourly_MERRA2"
    # meteo_vars = ['t2m', 'u2m', 'v2m', 'q2m', 'tpw', 'ps', 'slp']
    # data_type = ["mean"]
    # for Var in meteo_vars:
    #     print(Var)
    #     all_files2 = DownloadData(Dir, Var, Startdate, Enddate, latlim, lonlim, TimeStep, Period, username, password, Waitbar = True, data_type = data_type)




from osgeo import gdal
import tqdm
import os
import urllib
from pywapor.general.logger import log
from pywapor.general.pre_defaults import composite_defaults

def get_cog(url, out_file, bb, srs = "EPSG:4326"):

    # Check if out_file folder exists.
    folder = os.path.split(out_file)[0]
    if not os.path.isdir(folder):
        os.makedirs(folder)

    # Create waitbar
    waitbar = tqdm.tqdm(desc= f"Tile: 0 / 1",
                                position = 0,
                                total = 100,
                                bar_format='{l_bar}{bar}|')

    # Advance tile counter.
    waitbar_i = int(waitbar.desc.split(" ")[1])
    waitbar_desc = str(waitbar.desc)
    waitbar.set_description_str(waitbar_desc.replace(f": {waitbar_i} /", f": {waitbar_i+1} /")) 

    # Define callback function for waitbar progress.
    def callback_func(info, *args):
        waitbar.update(info * 100 - waitbar.n)

    # Set gdal.Translate options.
    options = gdal.TranslateOptions(
        projWin = bb,
        projWinSRS = srs,
        outputSRS = srs,
        callback = callback_func,
    )

    # Check if url is local or online path.
    is_not_local = urllib.parse.urlparse(url).scheme in ('http', 'https',)
    if is_not_local:
        url = f"/vsicurl/{url}"

    # Run gdal.Translate.
    ds = gdal.Translate(out_file, url, options = options)

    # Finish waitbar.
    waitbar.refresh()

    # Reset the gdal.Dataset.
    ds.FlushCache()
    ds = None

    return out_file

def collect(Dir, latlim, lonlim, vars = ['land_mask', 'lw_offset',
            'lw_slope', 'r0_bare', 'r0_full', 'rn_offset', 'rn_slope',
            'rs_min', 't_amp_year', 't_opt', 'vpd_slope', 'z_obst_max',
            'z_oro'], base_url = r"https://storage.googleapis.com/fao-cog-data", **kwargs):

    bb = [lonlim[0], latlim[1], lonlim[1], latlim[0]]

    files = list()

    log("--> Downloading STATICS.")

    cmetas = composite_defaults()

    for var in vars:

        var_name = var.replace("_", "-")

        if var in cmetas.keys():
            unit = cmetas[var]["var_unit"]
            fn = f"{var_name}_STATICS_{unit}_-_-.tif"
        else:
            fn = f"{var_name}_STATICS_unknown_-_-.tif"

        out_file = os.path.join(Dir, "STATICS", fn)
        url = os.path.join(base_url, f"L1_{var}.cog.tif")

        if not os.path.isfile(out_file):
            out_file = get_cog(url, out_file, bb)

        files.append(out_file)

    return files

if __name__ == "__main__":

    # url = r"http://landsat-pds.s3.amazonaws.com/c1/L8/073/087/LC08_L1TP_073087_20190818_20190902_01_T1/LC08_L1TP_073087_20190818_20190902_01_T1_B2.TIF"
    # url = r"/Volumes/Data/L1_STATICS_cogs/L1_land_mask.cog.tif"
    # bb = [175.2, -38.8, 175.8, -39.5]

    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]

    raw_folder = r"/Users/hmcoerver/On My Mac/cog_test"

    out = collect(raw_folder, latlim, lonlim)



# -*- coding: utf-8 -*-
import sys
from pywapor.collect.CHIRPS.DataAccess import DownloadData
import glob
import os
from pywapor.general.logger import log

def main(Dir, latlim, lonlim, Startdate, Enddate, Waitbar = 1):
	"""
	This function downloads daily CHIRPS data

	Keyword arguments:
	Dir -- 'C:/file/to/path/'
	Startdate -- 'yyyy-mm-dd'
	Enddate -- 'yyyy-mm-dd'
	latlim -- [ymin, ymax] (values must be between -50 and 50)
	lonlim -- [xmin, xmax] (values must be between -180 and 180)
	Waitbar -- 1 (Default) will print a waitbar
	"""

	log.info(f"--> Downloading CHIRPS.")
	# Download data
	DownloadData(Dir, Startdate, Enddate, latlim, lonlim, Waitbar)

	return glob.glob(os.path.join(Dir, "CHIRPS", "Precipitation", "*.tif"))

if __name__ == '__main__':
    main(sys.argv)



"""
Author: Laust Færch
Module: Collect/PROBAV
"""

import re
import os
import warnings
import rasterio
import rasterio.mask
import rasterio.merge
import numpy as np
import xarray as xr
from tqdm import tqdm
from pathlib import Path
from geojson import Polygon
from requests.exceptions import HTTPError
from datetime import datetime, timedelta
from aiohttp import ClientResponseError, ServerDisconnectedError
from asyncio import TimeoutError

# import importlib.util
# spec = importlib.util.spec_from_file_location("vito_download", "/Users/hmcoerver/Git/wapor-et-look/pywapor/collect/PROBAV/vito-download/vito_download/download.py")
# foo = importlib.util.module_from_spec(spec)
# spec.loader.exec_module(foo)
from pywapor.collect.PROBAV import download as foo

# Required for Python 3.6 and 3.7
import nest_asyncio
nest_asyncio.apply()

import vito_download as vito

# http://www.vito-eodata.be/PDF/datapool/Free_Data/PROBA-V_100m/S5_TOC_100_m_C1/2019/7/11/?coord=30.3,29.0,31.1,29.6
# /Users/hmcoerver/Git/wapor-et-look/tests/test_data/RAW/PROBAV/PROBAV_S5_TOC_X21Y04_20190711_100M_V101.HDF5

def download_data(download_dir, start_date, end_date, latitude_extent, longitude_extent, username,
                  password, buffer_dates = False):

    # setup
    max_retries = 5
    delete_hdf5 = False
    delete_tif = True
    dataset = 'Proba-V-S5-TOC'

    if not os.path.exists(download_dir):
        os.makedirs(download_dir)

    start_date = datetime.strptime(start_date, "%Y-%m-%d")
    end_date = datetime.strptime(end_date, "%Y-%m-%d")

    # Buffer the download to make sure that a decadal composite can be produced regarding of start
    # and end dates
    if buffer_dates:
        start_date = start_date - timedelta(days=5)
        end_date = end_date + timedelta(days=5)
    delta = end_date - start_date

    waitbar = tqdm(total = (delta.days)*10, position = 0, unit_scale = True)
    # Loop over all dates
    for i in range(delta.days):
        date = start_date + timedelta(days=i)
        waitbar.set_description_str(date.strftime("%Y.%m.%d:               "))

        template = os.path.join(download_dir, "{v}", "{v}_PROBAV_-_5-daily_{d_str}.tif")
        fh1 = template.format(v = "NDVI", d_str = date.strftime("%Y.%m.%d"))
        fh2 = template.format(v = "Albedo", d_str = date.strftime("%Y.%m.%d"))
        if np.all([os.path.isfile(fh1), os.path.isfile(fh2)]):
            waitbar.update(10)
            continue

        # retrieve vito URL
        url = vito.build_url(product=dataset, year=date.year, month=date.month, day=date.day,
                             extent={'xmin': longitude_extent[0], 'xmax': longitude_extent[1],
                                     'ymin': latitude_extent[0], 'ymax': latitude_extent[1]})

        no_of_attempts = 0
        download_success = False
        downloaded_files = []

        # sometimes vito.download fails, often it works to retry.
        while no_of_attempts <= max_retries:
            try:
                # download all matching files
                local_files = foo.download_data(url, username=username, password=password,
                                                 download_dir=download_dir, include='*.HDF5',
                                                 download_jobs=4)
                waitbar.set_description_str(date.strftime("%Y.%m.%d: Starting dl..."))
                downloaded_files = list(local_files)
                waitbar.set_description_str(date.strftime("%Y.%m.%d: dl finished   "))

                for file in downloaded_files:
                    # If file is corrupted, delete it and retry
                    try:
                        with xr.open_dataset(file, engine='netcdf4') as src:
                            crs = src.crs # all files should have crs
                    except OSError:
                        os.remove(file)
                        raise RuntimeError

                download_success = True

            except (RuntimeError, HTTPError,
                    ClientResponseError, ServerDisconnectedError, TimeoutError):
                no_of_attempts += 1
                continue
            break

        if not download_success:
            warnings.warn(f'vito download failed for date: {date.strftime("%Y-%m-%d")}')
            break

        if len(downloaded_files) == 0:
            waitbar.update(10)

        # convert downloaded HDF5 files to tif files
        for hdf_file in downloaded_files:
            waitbar.set_description_str(date.strftime("%Y.%m.%d: Converting to tif "))

            da = _hdf5_to_dataarray(hdf_file, 'LEVEL3/NDVI', 'NDVI')
            _dataarray_to_tif(da, str(Path(hdf_file).parent / Path(hdf_file).stem) + '_NDVI.tif')
            waitbar.update(1)

            band_list = ['BLUE', 'NIR', 'RED', 'SWIR']
            # read all bands and save as individual tifs
            for band in band_list:
                da = _hdf5_to_dataarray(hdf_file, f'LEVEL3/RADIOMETRY/{band}', 'TOC')
                _dataarray_to_tif(da, str(Path(hdf_file).parent /
                                          Path(hdf_file).stem) + f'_{band}.tif')
                waitbar.update(1)

            angle_band_list = ['VNIR', 'SWIR']
            # read all angle bands and save as individual tifs
            for band in angle_band_list:
                da = _hdf5_to_dataarray(hdf_file, f'LEVEL3/GEOMETRY/{band}', 'VZA')
                _dataarray_to_tif(da, str(Path(hdf_file).parent /
                                          Path(hdf_file).stem) + f'_{band}-VZA.tif')
                waitbar.update(1)

            # read and save quality mask
            da = _hdf5_to_dataarray(hdf_file, 'LEVEL3/QUALITY', 'SM')
            _dataarray_to_tif(da, str(Path(hdf_file).parent / Path(hdf_file).stem) + '_SM.tif')
            waitbar.update(1)

        # loop over all tif files in download folder
        input_files = []
        for tif_file in download_dir.glob('*.tif'):
            date_str = max(re.findall('[0-9]+', str(tif_file)), key=len)
            # collect all files that matches current date
            if date_str == date.strftime('%Y%m%d'):
                input_files.append(str(tif_file))

        for product in ["NDVI_PROBAV_-_5-daily_", "Albedo_PROBAV_-_5-daily_"]:
            waitbar.set_description_str(date.strftime("%Y.%m.%d: Saving finals.    "))

            output_file = str(download_dir /
                              ('%s/%s_%s.tif' % (product.split("_")[0], product[:-1], date.strftime("%Y.%m.%d"))))
            data_files = _preprocess_inputs(input_files, product)

            # merge files and clip to extent, save as tif
            if data_files:
                _merge_and_save_tifs(data_files, output_file, 
                                    latitude_extent, longitude_extent,
                                    delete_input=delete_tif)
                waitbar.update(1)

        if delete_tif:
            for file in input_files:
                os.remove(file)

    if delete_hdf5:
        for file in list(download_dir.glob('*.HDF5')):
            os.remove(file)

    waitbar.close()

    return()


# save xarray DataArrays as tif
def _dataarray_to_tif(da, filename):

    meta = {'driver': 'GTiff',
            'height': da.shape[0],
            'width': da.shape[1],
            'dtype': str(da.dtype),
            'count': 1,
            'transform': rasterio.Affine.from_gdal(*da.attrs['affine']),
            'crs': da.crs}

    with rasterio.open(filename, 'w', **meta) as dst:
        dst.write(da.values, 1)


# open hdf5 file as xarray DataArray
# if this function doesn't work: try to update your xarray and netcdf libraries.
def _hdf5_to_dataarray(filename, group, dataset_name):
    # read the group data
    with xr.open_dataset(filename, group=group, engine='netcdf4') as src:
        da = src[dataset_name]

    # fetch metadata
    with xr.open_dataset(filename) as src:
        meta = src.copy()

    da.attrs['crs'] = meta.crs.spatial_ref
    da.attrs['affine'] = [float(i) for i in meta.crs.GeoTransform.split(' ')[0:6]]

    return da


# merge tif files and save as one
def _merge_and_save_tifs(input_files, output_file, latitude_extent, longitude_extent,
                         delete_input=True):
    extent_poly = Polygon([(longitude_extent[0], latitude_extent[0]),
                           (longitude_extent[1], latitude_extent[0]),
                           (longitude_extent[1], latitude_extent[1]),
                           (longitude_extent[0], latitude_extent[1])])

    bbox = rasterio.features.bounds(extent_poly)
    src_files_to_mosaic = []
    crs = None

    for file in list(input_files):
        # use crs from first file as crs for mosaic
        if crs is None:
            with rasterio.open(file) as src:
                meta = src.profile
                crs = meta['crs']

        src_files_to_mosaic.append(rasterio.open(file))

    mosaic, mosaic_trans = rasterio.merge.merge(src_files_to_mosaic, bounds=bbox, target_aligned_pixels = True)
    mosaic = np.squeeze(mosaic)

    # Do not check if we have data in aoi


    meta = {
        'driver': 'GTiff',
        'width': mosaic.shape[1],
        'height': mosaic.shape[0],
        'count': 1,
        'dtype': str(mosaic.dtype),
        'crs': crs,
        'transform': mosaic_trans,
        'nodata': -9999
    }

    if not os.path.exists(os.path.dirname(output_file)):
        os.mkdir(os.path.dirname(output_file))
    with rasterio.open(output_file, 'w', **meta) as dst:
        dst.write(mosaic, 1)

    if delete_input:
        for src in src_files_to_mosaic:
            src.close()

        for file in input_files:
            os.remove(file)


# calculate albedo from band files
# Albedo should be calculated for each downloaded tile and saved as tif
def _preprocess_inputs(input_files, product):
    tiles = set([os.path.basename(file).split('_')[3] for file in input_files])
    updated_input_files = []

    for tile in list(tiles):
        matching_files = [elem for elem in input_files if tile in elem]

        all_data = []
        band_names = []
        for file in matching_files:
            with rasterio.open(str(file)) as src:
                all_data.append(src.read())
                meta = src.profile
            band_names.append(str(Path(file).stem).split('_')[-1])

        all_bands = np.asarray(all_data).squeeze()

        if product == "NDVI_PROBAV_-_5-daily_":
            data = all_bands[band_names.index('NDVI'), ...]
        elif product == "Albedo_PROBAV_-_5-daily_":
            data = (0.429 * all_bands[band_names.index('BLUE'), ...]
                    + 0.333 * all_bands[band_names.index('RED'), ...]
                    + 0.133 * all_bands[band_names.index('NIR'), ...]
                    + 0.105 * all_bands[band_names.index('SWIR'), ...])

        # Mask clouds
        quality_band = all_bands[band_names.index('SM'), ...]
        flag_mask = [1, 2, 3, 4]  # {1: shadow, 2: cloud, 3: undefined, 4: ice+snow}
        bit_mask_array = np.bitwise_or.reduce(flag_mask)*np.ones_like(quality_band)
        cloudmask = np.bitwise_and(quality_band.astype(np.int), bit_mask_array.astype(np.int)) > 0
        data[cloudmask] = None

        # Mask Viewing Angle
        max_angle = 35

        swir_vza = all_bands[band_names.index('SWIR-VZA'), ...]
        vnir_vza = all_bands[band_names.index('VNIR-VZA'), ...]

        swir_mask = swir_vza >= max_angle
        vnir_mask = vnir_vza >= max_angle

        if product == "NDVI_PROBAV_-_5-daily_":
            data[vnir_mask] = None
        elif product == "Albedo_PROBAV_-_5-daily_":
            data[vnir_mask & swir_mask] = None

        data_filename = str(Path(matching_files[0]).parent /
                            Path(str('_').join(Path(matching_files[0]).stem.split('_')[0:-1]))) + '_temp.tif'

        with rasterio.open(data_filename, 'w', **meta) as dst:
            dst.write(data, 1)

        updated_input_files.append(data_filename)

    return updated_input_files

# if __name__ == "__main__":

#     download_dir = Path(r"/Users/hmcoerver/On My Mac/probav_test/PROBAV")
#     latitude_extent = [28.9, 29.7]
#     longitude_extent = [30.2, 31.2]
#     start_date = "2021-07-01"
#     end_date = "2021-07-11"
#     username = ""
#     password = ""
#     buffer_dates = True

#     download_data(Path(download_dir), start_date, end_date, latitude_extent, longitude_extent, username,
#                   password, buffer_dates = True)

    # import warnings
    # warnings.simplefilter('always')

    # url = vito.build_url(product='Proba-V-S5-TOC', year=2021, month=7, day=6,
    #                     extent={'xmin': longitude_extent[0], 'xmax': longitude_extent[1],
    #                             'ymin': latitude_extent[0], 'ymax': latitude_extent[1]})

    # import importlib.util
    # spec = importlib.util.spec_from_file_location("vito_download", "/Users/hmcoerver/Git/wapor-et-look/pywapor/collect/PROBAV/vito-download/vito_download/download.py")
    # foo = importlib.util.module_from_spec(spec)
    # spec.loader.exec_module(foo)

    # local_files = foo.download_data(url, username=username, password=pw,
    #                                     download_dir=download_dir, include='*.HDF5',
    #                                     download_jobs=4)



    #%%


